@@ -197,7 +197,7 @@ char* gserialized_to_string(const GSERIALIZED *g)
 /* Unfortunately including advanced instructions is something that
 only helps a small sliver of users who can build their own
 knowing the target system they will be running on. Packagers
-have to aim for the lowest common demoninator. So this is
+have to aim for the lowest common denominator. So this is
 dead code for the forseeable future. */
 #define HAVE_PDEP 0
 #if HAVE_PDEP
@@ -300,35 +300,39 @@ int gserialized_cmp(const GSERIALIZED *g1, const GSERIALIZED *g2)
 	union floatuint x, y;
 
 	/*
-	* For two non-same planar points, we can skip a lot of machinery.
+	* For two non-same points, we can skip a lot of machinery.
 	*/
 	if (
 		sz1 > 16 && // 16 is size of EMPTY, if it's larger - it has coordinates
 		sz2 > 16 &&
-		*(uint32_t*)(g1->data) == POINTTYPE &&
-		*(uint32_t*)(g2->data) == POINTTYPE &&
 		!FLAGS_GET_BBOX(g1->flags) &&
-		!FLAGS_GET_GEODETIC(g1->flags) &&
 		!FLAGS_GET_BBOX(g2->flags) &&
-		!FLAGS_GET_GEODETIC(g2->flags)
+		*(uint32_t*)(g1+8) == POINTTYPE &&
+		*(uint32_t*)(g2+8) == POINTTYPE
 	)
 	{
-		double *dptr = (double*)(g1->data);
-		x.f = 2.0 * dptr[1];
-		y.f = 2.0 * dptr[2];
+		double *dptr = (double*)(g1->data + sizeof(double));
+		x.f = 2.0 * dptr[0];
+		y.f = 2.0 * dptr[1];
 		hash1 = uint32_interleave_2(x.u, y.u);
 
-		dptr = (double*)(g2->data);
-		x.f = 2.0 * dptr[1];
-		y.f = 2.0 * dptr[2];
+		dptr = (double*)(g2->data + sizeof(double));
+		x.f = 2.0 * dptr[0];
+		y.f = 2.0 * dptr[1];
 		hash2 = uint32_interleave_2(x.u, y.u);
 
-		if ( hash1 > hash2 )
-			return 1;
-		if ( hash1 < hash2 )
-			return -1;
+		/* If the SRIDs are the same, we can use hash inequality */
+		/* to jump us out of this function early. Otherwise we still */
+		/* have to do the full calculation */
+		if ( gserialized_cmp_srid(g1, g2) == 0 )
+		{
+			if ( hash1 > hash2 )
+				return 1;
+			if ( hash1 < hash2 )
+				return -1;
+		}
 
-		// if hashes happen to be the same, go to full compare.
+		/* if hashes happen to be the same, go to full compare. */
 	}
 
 	size_t hsz1 = gserialized_header_size(g1);
@@ -345,22 +349,12 @@ int gserialized_cmp(const GSERIALIZED *g1, const GSERIALIZED *g2)
 	g1_is_empty = (gserialized_get_gbox_p(g1, &box1) == LW_FAILURE);
 	g2_is_empty = (gserialized_get_gbox_p(g2, &box2) == LW_FAILURE);
 
-	/* Empty == Empty */
-	if (g1_is_empty && g2_is_empty)
-	{
-		/* POINT EMPTY == POINT EMPTY */
-		/* POINT EMPTY < LINESTRING EMPTY */
-		uint32_t t1 = gserialized_get_type(g1);
-		uint32_t t2 = gserialized_get_type(g2);
-		return t1 == t2 ? 0 : (t1 < t2 ? -1 : 1);
-	}
-
 	/* Empty < Non-empty */
-	if (g1_is_empty)
+	if (g1_is_empty && !g2_is_empty)
 		return -1;
 
 	/* Non-empty > Empty */
-	if (g2_is_empty)
+	if (!g1_is_empty && g2_is_empty)
 		return 1;
 
 	/* Return equality for perfect equality only */
@@ -368,48 +362,52 @@ int gserialized_cmp(const GSERIALIZED *g1, const GSERIALIZED *g2)
 	if ( bsz1 == bsz2 && cmp_srid == 0 && cmp == 0 )
 		return 0;
 
-	/* Using the centroids, calculate somewhat sortable */
-	/* hash key. The key doesn't provide good locality over */
-	/* the +/- boundary, but otherwise is pretty OK */
-	hash1 = gbox_get_sortable_hash(&box1);
-	hash2 = gbox_get_sortable_hash(&box2);
+	if (!g1_is_empty && !g2_is_empty)
+	{
+		/* Using the centroids, calculate somewhat sortable */
+		/* hash key. The key doesn't provide good locality over */
+		/* the +/- boundary, but otherwise is pretty OK */
+		hash1 = gbox_get_sortable_hash(&box1);
+		hash2 = gbox_get_sortable_hash(&box2);
 
-	if ( hash1 > hash2 )
-		return 1;
-	else if ( hash1 < hash2 )
-		return -1;
+		if ( hash1 > hash2 )
+			return 1;
+		else if ( hash1 < hash2 )
+			return -1;
 
-	/* What, the hashes are equal? OK... sort on the */
-	/* box minima */
-	if (box1.xmin < box2.xmin)
-		return -1;
-	else if (box1.xmin > box2.xmin)
-		return 1;
+		/* What, the hashes are equal? OK... sort on the */
+		/* box minima */
+		if (box1.xmin < box2.xmin)
+			return -1;
+		else if (box1.xmin > box2.xmin)
+			return 1;
 
-	if (box1.ymin < box2.ymin)
-		return -1;
-	else if (box1.ymin > box2.ymin)
-		return 1;
+		if (box1.ymin < box2.ymin)
+			return -1;
+		else if (box1.ymin > box2.ymin)
+			return 1;
 
-	/* Still equal? OK... sort on the box maxima */
-	if (box1.xmax < box2.xmax)
-		return -1;
-	else if (box1.xmax > box2.xmax)
-		return 1;
+		/* Still equal? OK... sort on the box maxima */
+		if (box1.xmax < box2.xmax)
+			return -1;
+		else if (box1.xmax > box2.xmax)
+			return 1;
 
-	if (box1.ymax < box2.ymax)
-		return -1;
-	else if (box1.ymax > box2.ymax)
-		return 1;
+		if (box1.ymax < box2.ymax)
+			return -1;
+		else if (box1.ymax > box2.ymax)
+			return 1;
+	}
 
-	/* Geeze! How about object size? Sort on that... */
-	if (hsz1 < hsz2)
-		return -1;
-	else if (hsz1 > hsz2)
-		return 1;
-
-	/* OK fine, we'll sort on the memcmp just to be done with this */
-	return cmp == 0 ? 0 : (cmp > 0 ? 1 : -1);
+	/* Prefix comes before longer one. */
+	if (bsz1 != bsz2 && cmp == 0)
+	{
+		if (bsz1 < bsz2)
+ 			return -1;
+		else if (bsz1 > bsz2)
+ 			return 1;
+	}
+	return cmp > 0 ? 1 : -1;
 }
 
 int gserialized_read_gbox_p(const GSERIALIZED *g, GBOX *gbox)
@@ -717,7 +715,7 @@ static size_t gserialized_from_lwtriangle_size(const LWTRIANGLE *triangle)
 static size_t gserialized_from_lwpoly_size(const LWPOLY *poly)
 {
 	size_t size = 4; /* Type number. */
-	int i = 0;
+	uint32_t i = 0;
 
 	assert(poly);
 
@@ -753,7 +751,7 @@ static size_t gserialized_from_lwcircstring_size(const LWCIRCSTRING *curve)
 static size_t gserialized_from_lwcollection_size(const LWCOLLECTION *col)
 {
 	size_t size = 4; /* Type number. */
-	int i = 0;
+	uint32_t i = 0;
 
 	assert(col);
 
@@ -904,7 +902,7 @@ static size_t gserialized_from_lwline(const LWLINE *line, uint8_t *buf)
 
 static size_t gserialized_from_lwpoly(const LWPOLY *poly, uint8_t *buf)
 {
-	int i;
+	uint32_t i;
 	uint8_t *loc;
 	int ptsize;
 	int type = POLYGONTYPE;
@@ -1036,7 +1034,7 @@ static size_t gserialized_from_lwcollection(const LWCOLLECTION *coll, uint8_t *b
 {
 	size_t subsize = 0;
 	uint8_t *loc;
-	int i;
+	uint32_t i;
 	int type;
 
 	assert(coll);
@@ -1195,6 +1193,8 @@ GSERIALIZED* gserialized_from_lwgeom(LWGEOM *geom, size_t *size)
 	*/
 	if ( geom->bbox )
 		FLAGS_SET_BBOX(geom->flags, 1);
+	else
+		FLAGS_SET_BBOX(geom->flags, 0);
 
 	/* Set up the uint8_t buffer into which we are going to write the serialized geometry. */
 	expected_size = gserialized_from_lwgeom_size(geom);
@@ -1314,7 +1314,7 @@ static LWPOLY* lwpoly_from_gserialized_buffer(uint8_t *data_ptr, uint8_t g_flags
 	LWPOLY *poly;
 	uint8_t *ordinate_ptr;
 	uint32_t nrings = 0;
-	int i = 0;
+	uint32_t i = 0;
 
 	assert(data_ptr);
 
@@ -1334,6 +1334,7 @@ static LWPOLY* lwpoly_from_gserialized_buffer(uint8_t *data_ptr, uint8_t g_flags
 	if ( nrings > 0)
 	{
 		poly->rings = (POINTARRAY**)lwalloc( sizeof(POINTARRAY*) * nrings );
+		poly->maxrings = nrings;
 		ordinate_ptr += nrings * 4; /* Move past all the npoints values. */
 		if ( nrings % 2 ) /* If there is padding, move past that too. */
 			ordinate_ptr += 4;
@@ -1341,6 +1342,7 @@ static LWPOLY* lwpoly_from_gserialized_buffer(uint8_t *data_ptr, uint8_t g_flags
 	else /* Empty polygon */
 	{
 		poly->rings = NULL;
+		poly->maxrings = 0;
 	}
 
 	for ( i = 0; i < nrings; i++ )
@@ -1431,7 +1433,7 @@ static LWCOLLECTION* lwcollection_from_gserialized_buffer(uint8_t *data_ptr, uin
 	uint8_t *start_ptr = data_ptr;
 	LWCOLLECTION *collection;
 	uint32_t ngeoms = 0;
-	int i = 0;
+	uint32_t i = 0;
 
 	assert(data_ptr);
 
@@ -1449,9 +1451,15 @@ static LWCOLLECTION* lwcollection_from_gserialized_buffer(uint8_t *data_ptr, uin
 	data_ptr += 4; /* Skip past the ngeoms. */
 
 	if ( ngeoms > 0 )
+	{
 		collection->geoms = lwalloc(sizeof(LWGEOM*) * ngeoms);
+		collection->maxgeoms = ngeoms;
+	}
 	else
+	{
 		collection->geoms = NULL;
+		collection->maxgeoms = 0;
+	}
 
 	/* Sub-geometries are never de-serialized with boxes (#1254) */
 	FLAGS_SET_BBOX(g_flags, 0);
@@ -1563,4 +1571,3 @@ LWGEOM* lwgeom_from_gserialized(const GSERIALIZED *g)
 
 	return lwgeom;
 }
-
