@@ -246,39 +246,6 @@ lwfree(void *mem)
 	lwfree_var(mem);
 }
 
-/*
- * Removes trailing zeros and dot for a %f formatted number.
- * Modifies input.
- */
-void
-trim_trailing_zeros(char *str)
-{
-	char *ptr, *totrim=NULL;
-	int len;
-	int i;
-
-	LWDEBUGF(3, "input: %s", str);
-
-	ptr = strchr(str, '.');
-	if ( ! ptr ) return; /* no dot, no decimal digits */
-
-	LWDEBUGF(3, "ptr: %s", ptr);
-
-	len = strlen(ptr);
-	for (i=len-1; i; i--)
-	{
-		if ( ptr[i] != '0' ) break;
-		totrim=&ptr[i];
-	}
-	if ( totrim )
-	{
-		if ( ptr == totrim-1 ) *ptr = '\0';
-		else *totrim = '\0';
-	}
-
-	LWDEBUGF(3, "output: %s", str);
-}
-
 /*
  * Returns a new string which contains a maximum of maxlength characters starting
  * from startpos and finishing at endpos (0-based indexing). If the string is
@@ -288,7 +255,7 @@ trim_trailing_zeros(char *str)
  * The caller should specify start or end truncation by setting the truncdirection
  * parameter as follows:
  *    0 - start truncation (i.e. characters are removed from the beginning)
- *    1 - end trunctation (i.e. characters are removed from the end)
+ *    1 - end truncation (i.e. characters are removed from the end)
  */
 
 char *lwmessage_truncate(char *str, int startpos, int endpos, int maxlength, int truncdirection)
@@ -359,7 +326,7 @@ char *lwmessage_truncate(char *str, int startpos, int endpos, int maxlength, int
 char
 getMachineEndian(void)
 {
-	static int endian_check_int = 1; /* dont modify this!!! */
+	static int endian_check_int = 1; /* don't modify this!!! */
 
 	return *((char *) &endian_check_int); /* 0 = big endian | xdr,
 	                                       * 1 = little endian | ndr
