@@ -50,7 +50,7 @@ ptarray_has_m(const POINTARRAY *pa)
  * Size of point represeneted in the POINTARRAY
  * 16 for 2d, 24 for 3d, 32 for 4d
  */
-int inline
+inline size_t
 ptarray_point_size(const POINTARRAY *pa)
 {
 	LWDEBUGF(5, "ptarray_point_size: FLAGS_NDIMS(pa->flags)=%x",FLAGS_NDIMS(pa->flags));
@@ -72,7 +72,7 @@ ptarray_construct_empty(char hasz, char hasm, uint32_t maxpoints)
 	POINTARRAY *pa = lwalloc(sizeof(POINTARRAY));
 	pa->serialized_pointlist = NULL;
 
-	/* Set our dimsionality info on the bitmap */
+	/* Set our dimensionality info on the bitmap */
 	pa->flags = gflags(hasz, hasm, 0);
 
 	/* We will be allocating a bit of room */
@@ -93,7 +93,7 @@ ptarray_construct_empty(char hasz, char hasm, uint32_t maxpoints)
 * pointarray supports.
 */
 int
-ptarray_insert_point(POINTARRAY *pa, const POINT4D *p, int where)
+ptarray_insert_point(POINTARRAY *pa, const POINT4D *p, uint32_t where)
 {
 	size_t point_size = ptarray_point_size(pa);
 	LWDEBUGF(5,"pa = %p; p = %p; where = %d", pa, p, where);
@@ -106,7 +106,7 @@ ptarray_insert_point(POINTARRAY *pa, const POINT4D *p, int where)
 	}
 
 	/* Error on invalid offset value */
-	if ( where > pa->npoints || where < 0)
+	if ( where > pa->npoints )
 	{
 		lwerror("ptarray_insert_point: offset out of range (%d)", where);
 		return LW_FAILURE;
@@ -123,7 +123,7 @@ ptarray_insert_point(POINTARRAY *pa, const POINT4D *p, int where)
 	/* Error out if we have a bad situation */
 	if ( pa->npoints > pa->maxpoints )
 	{
-		lwerror("npoints (%d) is greated than maxpoints (%d)", pa->npoints, pa->maxpoints);
+		lwerror("npoints (%d) is greater than maxpoints (%d)", pa->npoints, pa->maxpoints);
 		return LW_FAILURE;
 	}
 
@@ -258,10 +258,8 @@ ptarray_append_ptarray(POINTARRAY *pa1, POINTARRAY *pa2, double gap_tolerance)
 * pointarray supports.
 */
 int
-ptarray_remove_point(POINTARRAY *pa, int where)
+ptarray_remove_point(POINTARRAY *pa, uint32_t where)
 {
-	size_t ptsize = ptarray_point_size(pa);
-
 	/* Check for pathology */
 	if( ! pa )
 	{
@@ -270,17 +268,17 @@ ptarray_remove_point(POINTARRAY *pa, int where)
 	}
 
 	/* Error on invalid offset value */
-	if ( where >= pa->npoints || where < 0)
+	if ( where >= pa->npoints )
 	{
 		lwerror("ptarray_remove_point: offset out of range (%d)", where);
 		return LW_FAILURE;
 	}
 
 	/* If the point is any but the last, we need to copy the data back one point */
-	if( where < pa->npoints - 1 )
-	{
-		memmove(getPoint_internal(pa, where), getPoint_internal(pa, where+1), ptsize * (pa->npoints - where - 1));
-	}
+	if (where < pa->npoints - 1)
+		memmove(getPoint_internal(pa, where),
+			getPoint_internal(pa, where + 1),
+			ptarray_point_size(pa) * (pa->npoints - where - 1));
 
 	/* We have one less point */
 	pa->npoints--;
@@ -340,25 +338,26 @@ void ptarray_free(POINTARRAY *pa)
 
 
 void
-ptarray_reverse(POINTARRAY *pa)
+ptarray_reverse_in_place(POINTARRAY *pa)
 {
-	/* TODO change this to double array operations once point array is double aligned */
-	POINT4D pbuf;
-	uint32_t i;
-	int ptsize = ptarray_point_size(pa);
+	int i;
 	int last = pa->npoints-1;
 	int mid = pa->npoints/2;
 
-	for (i=0; i<mid; i++)
+	double *d = (double*)(pa->serialized_pointlist);
+	int j;
+	int ndims = FLAGS_NDIMS(pa->flags);
+	for (i = 0; i < mid; i++)
 	{
-		uint8_t *from, *to;
-		from = getPoint_internal(pa, i);
-		to = getPoint_internal(pa, (last-i));
-		memcpy((uint8_t *)&pbuf, to, ptsize);
-		memcpy(to, from, ptsize);
-		memcpy(from, (uint8_t *)&pbuf, ptsize);
+		for (j = 0; j < ndims; j++)
+		{
+			double buf;
+			buf = d[i*ndims+j];
+			d[i*ndims+j] = d[(last-i)*ndims+j];
+			d[(last-i)*ndims+j] = buf;
+		}
 	}
-
+	return;
 }
 
 
@@ -368,7 +367,7 @@ ptarray_reverse(POINTARRAY *pa)
 POINTARRAY*
 ptarray_flip_coordinates(POINTARRAY *pa)
 {
-	int i;
+	uint32_t i;
 	double d;
 	POINT4D p;
 
@@ -387,7 +386,7 @@ ptarray_flip_coordinates(POINTARRAY *pa)
 void
 ptarray_swap_ordinates(POINTARRAY *pa, LWORD o1, LWORD o2)
 {
-	int i;
+	uint32_t i;
 	double d, *dp1, *dp2;
 	POINT4D p;
 
@@ -418,7 +417,7 @@ ptarray_segmentize2d(const POINTARRAY *ipa, double dist)
 	POINT4D	p1, p2;
 	POINT4D pbuf;
 	POINTARRAY *opa;
-	int ipoff=0; /* input point offset */
+	uint32_t ipoff=0; /* input point offset */
 	int hasz = FLAGS_GET_Z(ipa->flags);
 	int hasm = FLAGS_GET_M(ipa->flags);
 
@@ -534,7 +533,6 @@ ptarray_addPoint(const POINTARRAY *pa, uint8_t *p, size_t pdims, uint32_t where)
 	ret = ptarray_construct(FLAGS_GET_Z(pa->flags),
 	                        FLAGS_GET_M(pa->flags), pa->npoints+1);
 
-	if ( where == -1 ) where = pa->npoints;
 
 	if ( where )
 	{
@@ -655,7 +653,7 @@ ptarray_clone(const POINTARRAY *in)
 {
 	POINTARRAY *out = lwalloc(sizeof(POINTARRAY));
 
-	LWDEBUG(3, "ptarray_clone_deep called.");
+	LWDEBUG(3, "ptarray_clone called.");
 
 	out->flags = in->flags;
 	out->npoints = in->npoints;
@@ -735,7 +733,7 @@ int
 ptarray_contains_point_partial(const POINTARRAY *pa, const POINT2D *pt, int check_closed, int *winding_number)
 {
 	int wn = 0;
-	int i;
+	uint32_t i;
 	double side;
 	const POINT2D *seg1;
 	const POINT2D *seg2;
@@ -781,7 +779,7 @@ ptarray_contains_point_partial(const POINTARRAY *pa, const POINT2D *pt, int chec
 		/*
 		* If the point is to the left of the line, and it's rising,
 		* then the line is to the right of the point and
-		* circling counter-clockwise, so incremement.
+		* circling counter-clockwise, so increment.
 		*/
 		if ( (side < 0) && (seg1->y <= pt->y) && (pt->y < seg2->y) )
 		{
@@ -834,7 +832,8 @@ int
 ptarrayarc_contains_point_partial(const POINTARRAY *pa, const POINT2D *pt, int check_closed, int *winding_number)
 {
 	int wn = 0;
-	int i, side;
+	uint32_t i;
+	int side;
 	const POINT2D *seg1;
 	const POINT2D *seg2;
 	const POINT2D *seg3;
@@ -995,7 +994,7 @@ ptarray_signed_area(const POINTARRAY *pa)
 	const POINT2D *P3;
 	double sum = 0.0;
 	double x0, x, y1, y2;
-	int i;
+	uint32_t i;
 
 	if (! pa || pa->npoints < 3 )
 		return 0.0;
@@ -1031,7 +1030,7 @@ POINTARRAY*
 ptarray_force_dims(const POINTARRAY *pa, int hasz, int hasm)
 {
 	/* TODO handle zero-length point arrays */
-	int i;
+	uint32_t i;
 	int in_hasz = FLAGS_GET_Z(pa->flags);
 	int in_hasm = FLAGS_GET_M(pa->flags);
 	POINT4D pt;
@@ -1298,9 +1297,9 @@ closest_point_on_segment(const POINT4D *p, const POINT4D *A, const POINT4D *B, P
 double
 ptarray_locate_point(const POINTARRAY *pa, const POINT4D *p4d, double *mindistout, POINT4D *proj4d)
 {
-	double mindist=-1;
+	double mindist=DBL_MAX;
 	double tlen, plen;
-	int t, seg=-1;
+	uint32_t t, seg=0;
 	POINT4D	start4d, end4d, projtmp;
 	POINT2D proj, p;
 	const POINT2D *start = NULL, *end = NULL;
@@ -1311,18 +1310,19 @@ ptarray_locate_point(const POINTARRAY *pa, const POINT4D *p4d, double *mindistou
 
 	if ( ! proj4d ) proj4d = &projtmp;
 
-	start = getPoint2d_cp(pa, 0);
-
-	/* If the pointarray has only one point, the nearest point is */
-	/* just that point */
-	if ( pa->npoints == 1 )
+	/* Check for special cases (length 0 and 1) */
+	if ( pa->npoints <= 1 )
 	{
-		getPoint4d_p(pa, 0, proj4d);
-		if ( mindistout )
-			*mindistout = distance2d_pt_pt(&p, start);
+		if ( pa->npoints == 1 )
+		{
+			getPoint4d_p(pa, 0, proj4d);
+			if ( mindistout )
+				*mindistout = distance2d_pt_pt(&p, getPoint2d_cp(pa, 0));
+		}
 		return 0.0;
 	}
 
+	start = getPoint2d_cp(pa, 0);
 	/* Loop through pointarray looking for nearest segment */
 	for (t=1; t<pa->npoints; t++)
 	{
@@ -1330,16 +1330,15 @@ ptarray_locate_point(const POINTARRAY *pa, const POINT4D *p4d, double *mindistou
 		end = getPoint2d_cp(pa, t);
 		dist = distance2d_pt_seg(&p, start, end);
 
-		if (t==1 || dist < mindist )
+		if ( dist < mindist )
 		{
 			mindist = dist;
 			seg=t-1;
-		}
-
-		if ( mindist == 0 )
-		{
-			LWDEBUG(3, "Breaking on mindist=0");
-			break;
+			if ( mindist == 0 )
+			{
+				LWDEBUG(3, "Breaking on mindist=0");
+				break;
+			}
 		}
 
 		start = end;
@@ -1409,7 +1408,7 @@ ptarray_locate_point(const POINTARRAY *pa, const POINT4D *p4d, double *mindistou
 void
 ptarray_longitude_shift(POINTARRAY *pa)
 {
-	int i;
+	uint32_t i;
 	double x;
 
 	for (i=0; i<pa->npoints; i++)
@@ -1427,67 +1426,12 @@ ptarray_longitude_shift(POINTARRAY *pa)
  * removed. Equality test on all dimensions of input.
  *
  * Always returns a newly allocated object.
- *
  */
-POINTARRAY *
+static POINTARRAY *
 ptarray_remove_repeated_points_minpoints(const POINTARRAY *in, double tolerance, int minpoints)
 {
-	POINTARRAY* out;
-	size_t ptsize;
-	size_t ipn, opn;
-	const POINT2D *last_point, *this_point;
-	double tolsq = tolerance * tolerance;
-
-	if ( minpoints < 1 ) minpoints = 1;
-
-	LWDEBUGF(3, "%s called", __func__);
-
-	/* Single or zero point arrays can't have duplicates */
-	if ( in->npoints < 3 ) return ptarray_clone_deep(in);
-
-	ptsize = ptarray_point_size(in);
-
-	LWDEBUGF(3, " ptsize: %d", ptsize);
-
-	/* Allocate enough space for all points */
-	out = ptarray_construct(FLAGS_GET_Z(in->flags),
-	                        FLAGS_GET_M(in->flags), in->npoints);
-
-	/* Now fill up the actual points (NOTE: could be optimized) */
-
-	opn=1;
-	/* Keep the first point */
-	memcpy(getPoint_internal(out, 0), getPoint_internal(in, 0), ptsize);
-	last_point = getPoint2d_cp(in, 0);
-	LWDEBUGF(3, " first point copied, out points: %d", opn);
-	for ( ipn = 1; ipn < in->npoints; ++ipn)
-	{
-		this_point = getPoint2d_cp(in, ipn);
-		if ( ipn < in->npoints-minpoints+1 || opn >= minpoints ) /* need extra points to hit minponts */
-		{
-			if (
-				(tolerance == 0 && memcmp(getPoint_internal(in, ipn-1), getPoint_internal(in, ipn), ptsize) == 0) || /* exact dupe */
-				(tolerance > 0.0 && distance2d_sqr_pt_pt(last_point, this_point) <= tolsq) /* within the removal tolerance */
-			) continue;
-		}
-
-		/*
-		 * The point is different (see above) from the previous,
-		 * so we add it to output
-		 */
-		memcpy(getPoint_internal(out, opn++), getPoint_internal(in, ipn), ptsize);
-		last_point = this_point;
-		LWDEBUGF(3, " Point %d differs from point %d. Out points: %d", ipn, ipn-1, opn);
-	}
-	/* Keep the last point */
-	if ( memcmp(last_point, getPoint_internal(in, ipn-1), ptsize) != 0 )
-	{
-		memcpy(getPoint_internal(out, opn-1), getPoint_internal(in, ipn-1), ptsize);
-	}
-
-	LWDEBUGF(3, " in:%d out:%d", out->npoints, opn);
-	out->npoints = opn;
-
+	POINTARRAY *out = ptarray_clone_deep(in);
+	ptarray_remove_repeated_points_in_place(out, tolerance, minpoints);
 	return out;
 }
 
@@ -1497,8 +1441,75 @@ ptarray_remove_repeated_points(const POINTARRAY *in, double tolerance)
 	return ptarray_remove_repeated_points_minpoints(in, tolerance, 2);
 }
 
+
+void
+ptarray_remove_repeated_points_in_place(POINTARRAY *pa, double tolerance, uint32_t min_points)
+{
+	uint32_t i;
+	double tolsq = tolerance * tolerance;
+	const POINT2D *last = NULL;
+	const POINT2D *pt;
+	uint32_t n_points = pa->npoints;
+	uint32_t n_points_out = 1;
+	size_t pt_size = ptarray_point_size(pa);
+
+	double dsq = FLT_MAX;
+
+	/* No-op on short inputs */
+	if ( n_points <= min_points ) return;
+
+	last = getPoint2d_cp(pa, 0);
+	for (i = 1; i < n_points; i++)
+	{
+		int last_point = (i == n_points-1);
+
+		/* Look straight into the abyss */
+		pt = getPoint2d_cp(pa, i);
+
+		/* Don't drop points if we are running short of points */
+		if (n_points + n_points_out > min_points + i)
+		{
+			if (tolerance > 0.0)
+			{
+				/* Only drop points that are within our tolerance */
+				dsq = distance2d_sqr_pt_pt(last, pt);
+				/* Allow any point but the last one to be dropped */
+				if (!last_point && dsq <= tolsq)
+				{
+					continue;
+				}
+			}
+			else
+			{
+				/* At tolerance zero, only skip exact dupes */
+				if (memcmp((char*)pt, (char*)last, pt_size) == 0)
+					continue;
+			}
+
+			/* Got to last point, and it's not very different from */
+			/* the point that preceded it. We want to keep the last */
+			/* point, not the second-to-last one, so we pull our write */
+			/* index back one value */
+			if (last_point && n_points_out > 1 && tolerance > 0.0 && dsq <= tolsq)
+			{
+				n_points_out--;
+			}
+		}
+
+		/* Compact all remaining values to front of array */
+		ptarray_copy_point(pa, i, n_points_out++);
+		last = pt;
+	}
+	/* Adjust array length */
+	pa->npoints = n_points_out;
+	return;
+}
+
+
+/************************************************************************/
+
 static void
-ptarray_dp_findsplit(POINTARRAY *pts, int p1, int p2, int *split, double *dist)
+ptarray_dp_findsplit_in_place(const POINTARRAY *pts, int p1, int p2, int *split, double *dist)
 {
 	int k;
 	const POINT2D *pk, *pa, *pb;
@@ -1536,75 +1547,105 @@ ptarray_dp_findsplit(POINTARRAY *pts, int p1, int p2, int *split, double *dist)
 			}
 		}
 		*dist = d;
-
-	} /* length---should be redone if can == 0 */
+	}
 	else
 	{
 		LWDEBUG(3, "segment too short, no split/no dist");
 		*dist = -1;
 	}
-
 }
 
-POINTARRAY *
-ptarray_simplify(POINTARRAY *inpts, double epsilon, unsigned int minpts)
+static int
+int_cmp(const void *a, const void *b)
 {
-	int *stack;			/* recursion stack */
-	int sp=-1;			/* recursion stack pointer */
-	int p1, split;
-	double dist;
-	POINTARRAY *outpts;
-	POINT4D pt;
+	/* casting pointer types */
+	const int *ia = (const int *)a;
+	const int *ib = (const int *)b;
+	/* returns negative if b > a and positive if a > b */
+	return *ia - *ib;
+}
 
+void
+ptarray_simplify_in_place(POINTARRAY *pa, double epsilon, uint32_t minpts)
+{
+	static size_t stack_size = 256;
+	int *stack, *outlist; /* recursion stack */
+	int stack_static[stack_size];
+	int outlist_static[stack_size];
+	int sp = -1; /* recursion stack pointer */
+	int p1, split;
+	uint32_t outn = 0;
+	int pai = 0;
+	uint32_t i;
+	double dist;
 	double eps_sqr = epsilon * epsilon;
 
-	/* Allocate recursion stack */
-	stack = lwalloc(sizeof(int)*inpts->npoints);
+	/* Do not try to simplify really short things */
+	if (pa->npoints < 3) return;
+
+	/* Only heap allocate book-keeping arrays if necessary */
+	if (pa->npoints > stack_size)
+	{
+		stack = lwalloc(sizeof(int) * pa->npoints);
+		outlist = lwalloc(sizeof(int) * pa->npoints);
+	}
+	else
+	{
+		stack = stack_static;
+		outlist = outlist_static;
+	}
 
 	p1 = 0;
-	stack[++sp] = inpts->npoints-1;
-
-	LWDEBUGF(2, "Input has %d pts and %d dims", inpts->npoints,
-	                                            FLAGS_NDIMS(inpts->flags));
-
-	/* Allocate output POINTARRAY, and add first point. */
-	outpts = ptarray_construct_empty(FLAGS_GET_Z(inpts->flags), FLAGS_GET_M(inpts->flags), inpts->npoints);
-	getPoint4d_p(inpts, 0, &pt);
-	ptarray_append_point(outpts, &pt, LW_FALSE);
-
-	LWDEBUG(3, "Added P0 to simplified point array (size 1)");
+	stack[++sp] = pa->npoints-1;
 
+	/* Add first point to output list */
+	outlist[outn++] = 0;
 	do
 	{
+		ptarray_dp_findsplit_in_place(pa, p1, stack[sp], &split, &dist);
 
-		ptarray_dp_findsplit(inpts, p1, stack[sp], &split, &dist);
-
-		LWDEBUGF(3, "Farthest point from P%d-P%d is P%d (dist. %g)", p1, stack[sp], split, dist);
-
-		if (dist > eps_sqr || ( outpts->npoints+sp+1 < minpts && dist >= 0 ) )
+		if ((dist > eps_sqr) || ((outn + sp+1 < minpts) && (dist >= 0)))
 		{
-			LWDEBUGF(4, "Added P%d to stack (outpts:%d)", split, sp);
 			stack[++sp] = split;
 		}
 		else
 		{
-			getPoint4d_p(inpts, stack[sp], &pt);
-			LWDEBUGF(4, "npoints , minpoints %d %d", outpts->npoints, minpts);
-			ptarray_append_point(outpts, &pt, LW_FALSE);
-
-			LWDEBUGF(4, "Added P%d to simplified point array (size: %d)", stack[sp], outpts->npoints);
-
+			outlist[outn++] = stack[sp];
 			p1 = stack[sp--];
 		}
-
-		LWDEBUGF(4, "stack pointer = %d", sp);
 	}
-	while (! (sp<0) );
+	while (!(sp<0));
 
-	lwfree(stack);
-	return outpts;
+	/* Put list of retained points into order */
+	qsort(outlist, outn, sizeof(int), int_cmp);
+	/* Copy retained points to front of array */
+	for (i = 0; i < outn; i++)
+	{
+		int j = outlist[i];
+		/* Indexes the same, means no copy required */
+		if (j == pai)
+		{
+			pai++;
+			continue;
+		}
+		/* Indexes different, copy value down */
+		ptarray_copy_point(pa, j, pai++);
+	}
+
+	/* Adjust point count on array */
+	pa->npoints = outn;
+
+	/* Only free if arrays are on heap */
+	if (stack != stack_static)
+		lwfree(stack);
+	if (outlist != outlist_static)
+		lwfree(outlist);
+
+	return;
 }
 
+/************************************************************************/
+
 /**
 * Find the 2d length of the given #POINTARRAY, using circular
 * arc interpolation between each coordinate triple.
@@ -1614,19 +1655,19 @@ double
 ptarray_arc_length_2d(const POINTARRAY *pts)
 {
 	double dist = 0.0;
-	int i;
+	uint32_t i;
 	const POINT2D *a1;
 	const POINT2D *a2;
 	const POINT2D *a3;
 
 	if ( pts->npoints % 2 != 1 )
-        lwerror("arc point array with even number of points");
+		lwerror("arc point array with even number of points");
 
 	a1 = getPoint2d_cp(pts, 0);
 
 	for ( i=2; i < pts->npoints; i += 2 )
 	{
-    	a2 = getPoint2d_cp(pts, i-1);
+		a2 = getPoint2d_cp(pts, i-1);
 		a3 = getPoint2d_cp(pts, i);
 		dist += lw_arc_length(a1, a2, a3);
 		a1 = a3;
@@ -1641,7 +1682,7 @@ double
 ptarray_length_2d(const POINTARRAY *pts)
 {
 	double dist = 0.0;
-	int i;
+	uint32_t i;
 	const POINT2D *frm;
 	const POINT2D *to;
 
@@ -1669,7 +1710,7 @@ double
 ptarray_length(const POINTARRAY *pts)
 {
 	double dist = 0.0;
-	int i;
+	uint32_t i;
 	POINT3DZ frm;
 	POINT3DZ to;
 
@@ -1699,7 +1740,7 @@ ptarray_length(const POINTARRAY *pts)
  * aligned to double boundary.
  */
 uint8_t *
-getPoint_internal(const POINTARRAY *pa, int n)
+getPoint_internal(const POINTARRAY *pa, uint32_t n)
 {
 	size_t size;
 	uint8_t *ptr;
@@ -1713,8 +1754,7 @@ getPoint_internal(const POINTARRAY *pa, int n)
 
 	LWDEBUGF(5, "(n=%d, pa.npoints=%d, pa.maxpoints=%d)",n,pa->npoints,pa->maxpoints);
 
-	if ( ( n < 0 ) ||
-	     ( n > pa->npoints ) ||
+	if ( ( n > pa->npoints ) ||
 	     ( n >= pa->maxpoints ) )
 	{
 		lwerror("%s [%d] called outside of ptarray range (n=%d, pa.npoints=%d, pa.maxpoints=%d)", __FILE__, __LINE__, n, pa->npoints, pa->maxpoints);
@@ -1748,7 +1788,7 @@ getPoint_internal(const POINTARRAY *pa, int n)
 void
 ptarray_affine(POINTARRAY *pa, const AFFINE *a)
 {
-	int i;
+	uint32_t i;
 	double x,y,z;
 	POINT4D p4d;
 
@@ -1793,40 +1833,170 @@ ptarray_affine(POINTARRAY *pa, const AFFINE *a)
 
 }
 
+/**
+* WARNING, make sure you send in only 16-member double arrays
+* or obviously things will go pear-shaped fast.
+*/
+#if 0
+static int gluInvertMatrix(const double *m, double *invOut)
+{
+    double inv[16], det;
+    int i;
+
+    inv[0] = m[5]  * m[10] * m[15] -
+             m[5]  * m[11] * m[14] -
+             m[9]  * m[6]  * m[15] +
+             m[9]  * m[7]  * m[14] +
+             m[13] * m[6]  * m[11] -
+             m[13] * m[7]  * m[10];
+
+    inv[4] = -m[4]  * m[10] * m[15] +
+              m[4]  * m[11] * m[14] +
+              m[8]  * m[6]  * m[15] -
+              m[8]  * m[7]  * m[14] -
+              m[12] * m[6]  * m[11] +
+              m[12] * m[7]  * m[10];
+
+    inv[8] = m[4]  * m[9] * m[15] -
+             m[4]  * m[11] * m[13] -
+             m[8]  * m[5] * m[15] +
+             m[8]  * m[7] * m[13] +
+             m[12] * m[5] * m[11] -
+             m[12] * m[7] * m[9];
+
+    inv[12] = -m[4]  * m[9] * m[14] +
+               m[4]  * m[10] * m[13] +
+               m[8]  * m[5] * m[14] -
+               m[8]  * m[6] * m[13] -
+               m[12] * m[5] * m[10] +
+               m[12] * m[6] * m[9];
+
+    inv[1] = -m[1]  * m[10] * m[15] +
+              m[1]  * m[11] * m[14] +
+              m[9]  * m[2] * m[15] -
+              m[9]  * m[3] * m[14] -
+              m[13] * m[2] * m[11] +
+              m[13] * m[3] * m[10];
+
+    inv[5] = m[0]  * m[10] * m[15] -
+             m[0]  * m[11] * m[14] -
+             m[8]  * m[2] * m[15] +
+             m[8]  * m[3] * m[14] +
+             m[12] * m[2] * m[11] -
+             m[12] * m[3] * m[10];
+
+    inv[9] = -m[0]  * m[9] * m[15] +
+              m[0]  * m[11] * m[13] +
+              m[8]  * m[1] * m[15] -
+              m[8]  * m[3] * m[13] -
+              m[12] * m[1] * m[11] +
+              m[12] * m[3] * m[9];
+
+    inv[13] = m[0]  * m[9] * m[14] -
+              m[0]  * m[10] * m[13] -
+              m[8]  * m[1] * m[14] +
+              m[8]  * m[2] * m[13] +
+              m[12] * m[1] * m[10] -
+              m[12] * m[2] * m[9];
+
+    inv[2] = m[1]  * m[6] * m[15] -
+             m[1]  * m[7] * m[14] -
+             m[5]  * m[2] * m[15] +
+             m[5]  * m[3] * m[14] +
+             m[13] * m[2] * m[7] -
+             m[13] * m[3] * m[6];
+
+    inv[6] = -m[0]  * m[6] * m[15] +
+              m[0]  * m[7] * m[14] +
+              m[4]  * m[2] * m[15] -
+              m[4]  * m[3] * m[14] -
+              m[12] * m[2] * m[7] +
+              m[12] * m[3] * m[6];
+
+    inv[10] = m[0]  * m[5] * m[15] -
+              m[0]  * m[7] * m[13] -
+              m[4]  * m[1] * m[15] +
+              m[4]  * m[3] * m[13] +
+              m[12] * m[1] * m[7] -
+              m[12] * m[3] * m[5];
+
+    inv[14] = -m[0]  * m[5] * m[14] +
+               m[0]  * m[6] * m[13] +
+               m[4]  * m[1] * m[14] -
+               m[4]  * m[2] * m[13] -
+               m[12] * m[1] * m[6] +
+               m[12] * m[2] * m[5];
+
+    inv[3] = -m[1] * m[6] * m[11] +
+              m[1] * m[7] * m[10] +
+              m[5] * m[2] * m[11] -
+              m[5] * m[3] * m[10] -
+              m[9] * m[2] * m[7] +
+              m[9] * m[3] * m[6];
+
+    inv[7] = m[0] * m[6] * m[11] -
+             m[0] * m[7] * m[10] -
+             m[4] * m[2] * m[11] +
+             m[4] * m[3] * m[10] +
+             m[8] * m[2] * m[7] -
+             m[8] * m[3] * m[6];
+
+    inv[11] = -m[0] * m[5] * m[11] +
+               m[0] * m[7] * m[9] +
+               m[4] * m[1] * m[11] -
+               m[4] * m[3] * m[9] -
+               m[8] * m[1] * m[7] +
+               m[8] * m[3] * m[5];
+
+    inv[15] = m[0] * m[5] * m[10] -
+              m[0] * m[6] * m[9] -
+              m[4] * m[1] * m[10] +
+              m[4] * m[2] * m[9] +
+              m[8] * m[1] * m[6] -
+              m[8] * m[2] * m[5];
+
+    det = m[0] * inv[0] + m[1] * inv[4] + m[2] * inv[8] + m[3] * inv[12];
+
+    if (det == 0)
+        return LW_FALSE;
+
+    det = 1.0 / det;
+
+    for (i = 0; i < 16; i++)
+        invOut[i] = inv[i] * det;
+
+    return LW_TRUE;
+}
+#endif
+
 /**
  * Scale a pointarray.
  */
 void
 ptarray_scale(POINTARRAY *pa, const POINT4D *fact)
 {
-  int i;
-  POINT4D p4d;
-
-  LWDEBUG(3, "ptarray_scale start");
-
-  for (i=0; i<pa->npoints; ++i)
-  {
-    getPoint4d_p(pa, i, &p4d);
-    p4d.x *= fact->x;
-    p4d.y *= fact->y;
-    p4d.z *= fact->z;
-    p4d.m *= fact->m;
-    ptarray_set_point4d(pa, i, &p4d);
-  }
-
-  LWDEBUG(3, "ptarray_scale end");
-
+	uint32_t i;
+	POINT4D p4d;
+	LWDEBUG(3, "ptarray_scale start");
+	for (i=0; i<pa->npoints; i++)
+	{
+		getPoint4d_p(pa, i, &p4d);
+		p4d.x *= fact->x;
+		p4d.y *= fact->y;
+		p4d.z *= fact->z;
+		p4d.m *= fact->m;
+		ptarray_set_point4d(pa, i, &p4d);
+	}
+	LWDEBUG(3, "ptarray_scale end");
 }
 
 int
-ptarray_startpoint(const POINTARRAY* pa, POINT4D* pt)
+ptarray_startpoint(const POINTARRAY *pa, POINT4D *pt)
 {
 	return getPoint4d_p(pa, 0, pt);
 }
 
 
-
-
 /*
  * Stick an array of points to the given gridspec.
  * Return "gridded" points in *outpts and their number in *outptsn.
@@ -1835,51 +2005,80 @@ ptarray_startpoint(const POINTARRAY* pa, POINT4D* pt)
  * into one single point.
  *
  */
-POINTARRAY *
-ptarray_grid(const POINTARRAY *pa, const gridspec *grid)
+void
+ptarray_grid_in_place(POINTARRAY *pa, const gridspec *grid)
 {
-	POINT4D pt;
-	int ipn; /* input point numbers */
-	POINTARRAY *dpa;
+	uint32_t i, j = 0;
+	POINT4D *p, *p_out = NULL;
+	int ndims = FLAGS_NDIMS(pa->flags);
+	int has_z = FLAGS_GET_Z(pa->flags);
+	int has_m = FLAGS_GET_M(pa->flags);
 
-	LWDEBUGF(2, "ptarray_grid called on %p", pa);
+	LWDEBUGF(2, "%s called on %p", __func__, pa);
 
-	dpa = ptarray_construct_empty(FLAGS_GET_Z(pa->flags),FLAGS_GET_M(pa->flags), pa->npoints);
-
-	for (ipn=0; ipn<pa->npoints; ++ipn)
+	for (i = 0; i < pa->npoints; i++)
 	{
+		/* Look straight into the abyss */
+		p = (POINT4D*)(getPoint_internal(pa, i));
 
-		getPoint4d_p(pa, ipn, &pt);
+		if (grid->xsize > 0)
+		{
+			p->x = rint((p->x - grid->ipx)/grid->xsize) * grid->xsize + grid->ipx;
+		}
 
-		if ( grid->xsize )
-			pt.x = rint((pt.x - grid->ipx)/grid->xsize) *
-			         grid->xsize + grid->ipx;
+		if (grid->ysize > 0)
+		{
+			p->y = rint((p->y - grid->ipy)/grid->ysize) * grid->ysize + grid->ipy;
+		}
 
-		if ( grid->ysize )
-			pt.y = rint((pt.y - grid->ipy)/grid->ysize) *
-			         grid->ysize + grid->ipy;
+		/* Read and round this point */
+		/* Z is always in third position */
+		if (has_z)
+		{
+			if (grid->zsize > 0)
+				p->z = rint((p->z - grid->ipz)/grid->zsize) * grid->zsize + grid->ipz;
+		}
+		/* M might be in 3rd or 4th position */
+		if (has_m)
+		{
+			/* In POINT M, M is in 3rd position */
+			if (grid->msize > 0 && !has_z)
+				p->z = rint((p->z - grid->ipm)/grid->msize) * grid->msize + grid->ipm;
+			/* In POINT ZM, M is in 4th position */
+			if (grid->msize > 0 && has_z)
+				p->m = rint((p->m - grid->ipm)/grid->msize) * grid->msize + grid->ipm;
+		}
 
-		if ( FLAGS_GET_Z(pa->flags) && grid->zsize )
-			pt.z = rint((pt.z - grid->ipz)/grid->zsize) *
-			         grid->zsize + grid->ipz;
-
-		if ( FLAGS_GET_M(pa->flags) && grid->msize )
-			pt.m = rint((pt.m - grid->ipm)/grid->msize) *
-			         grid->msize + grid->ipm;
-
-		ptarray_append_point(dpa, &pt, LW_FALSE);
+		/* Skip duplicates */
+		if ( p_out && FP_EQUALS(p_out->x, p->x) && FP_EQUALS(p_out->y, p->y)
+		   && (ndims > 2 ? FP_EQUALS(p_out->z, p->z) : 1)
+		   && (ndims > 3 ? FP_EQUALS(p_out->m, p->m) : 1) )
+		{
+			continue;
+		}
 
+		/* Write rounded values into the next available point */
+		p_out = (POINT4D*)(getPoint_internal(pa, j++));
+		p_out->x = p->x;
+		p_out->y = p->y;
+		if (ndims > 2)
+			p_out->z = p->z;
+		if (ndims > 3)
+			p_out->m = p->m;
 	}
 
-	return dpa;
+	/* Update output ptarray length */
+	pa->npoints = j;
+	return;
 }
 
+
 int
 ptarray_npoints_in_rect(const POINTARRAY *pa, const GBOX *gbox)
 {
 	const POINT2D *pt;
 	int n = 0;
-	int i;
+	uint32_t i;
 	for ( i = 0; i < pa->npoints; i++ )
 	{
 		pt = getPoint2d_cp(pa, i);
