@@ -26,6 +26,7 @@
 #define _LWGEOM_RTREE_H 1
 
 #include "liblwgeom.h"
+#include "lwgeom_cache.h"
 
 /**
 * Representation for the y-axis interval spanned by an edge.
@@ -59,22 +60,15 @@ typedef struct
 	RTREE_NODE **ringIndices;
 	int* ringCounts;
 	int polyCount;
-}
-RTREE_POLY_CACHE;
+} RTREE_POLY_CACHE;
 
 
-
-typedef struct {
-	int                         type;       // <GeomCache>
-	GSERIALIZED*                geom1;      //
-	GSERIALIZED*                geom2;      //
-	size_t                      geom1_size; //
-	size_t                      geom2_size; //
-	int32                       argnum;     // </GeomCache>
-	RTREE_POLY_CACHE*           index;
+typedef struct
+{
+	GeomCache             gcache;
+	RTREE_POLY_CACHE      *index;
 } RTreeGeomCache;
 
-
 /**
 * Retrieves a collection of line segments given the root and crossing value.
 */
