@@ -39,55 +39,64 @@
 
 #define FEATURES_CAPACITY_INITIAL 50
 
-enum mvt_cmd_id {
+enum mvt_cmd_id
+{
 	CMD_MOVE_TO = 1,
 	CMD_LINE_TO = 2,
 	CMD_CLOSE_PATH = 7
 };
 
-enum mvt_type {
+enum mvt_type
+{
 	MVT_POINT = 1,
 	MVT_LINE = 2,
 	MVT_RING = 3
 };
 
-struct mvt_kv_key {
+struct mvt_kv_key
+{
 	char *name;
 	uint32_t id;
 	UT_hash_handle hh;
 };
 
-struct mvt_kv_string_value {
+struct mvt_kv_string_value
+{
 	char *string_value;
 	uint32_t id;
 	UT_hash_handle hh;
 };
 
-struct mvt_kv_float_value {
+struct mvt_kv_float_value
+{
 	float float_value;
 	uint32_t id;
 	UT_hash_handle hh;
 };
 
-struct mvt_kv_double_value {
+struct mvt_kv_double_value
+{
 	double double_value;
 	uint32_t id;
 	UT_hash_handle hh;
 };
 
-struct mvt_kv_uint_value {
+struct mvt_kv_uint_value
+{
 	uint64_t uint_value;
 	uint32_t id;
 	UT_hash_handle hh;
 };
 
-struct mvt_kv_sint_value {
+struct mvt_kv_sint_value
+{
 	int64_t sint_value;
 	uint32_t id;
 	UT_hash_handle hh;
 };
 
-struct mvt_kv_bool_value {
+struct mvt_kv_bool_value
+{
 	protobuf_c_boolean bool_value;
 	uint32_t id;
 	UT_hash_handle hh;
@@ -103,23 +112,25 @@ static inline uint32_t p_int(int32_t value)
 	return (value << 1) ^ (value >> 31);
 }
 
-static uint32_t encode_ptarray(struct mvt_agg_context *ctx, enum mvt_type type,
-			       POINTARRAY *pa, uint32_t *buffer,
+static uint32_t encode_ptarray(__attribute__((__unused__)) mvt_agg_context *ctx,
+			       enum mvt_type type, POINTARRAY *pa, uint32_t *buffer,
 			       int32_t *px, int32_t *py)
 {
 	uint32_t offset = 0;
 	uint32_t i, c = 0;
 	int32_t dx, dy, x, y;
+	const POINT2D *p;
 
 	/* loop points and add to buffer */
-	for (i = 0; i < pa->npoints; i++) {
+	for (i = 0; i < pa->npoints; i++)
+	{
 		/* move offset for command */
 		if (i == 0 || (i == 1 && type > MVT_POINT))
 			offset++;
 		/* skip closing point for rings */
 		if (type == MVT_RING && i == pa->npoints - 1)
 			break;
-		const POINT2D *p = getPoint2d_cp(pa, i);
+		p = getPoint2d_cp(pa, i);
 		x = p->x;
 		y = p->y;
 		dx = x - *px;
@@ -132,10 +143,13 @@ static uint32_t encode_ptarray(struct mvt_agg_context *ctx, enum mvt_type type,
 	}
 
 	/* determine initial move and eventual line command */
-	if (type == MVT_POINT) {
+	if (type == MVT_POINT)
+	{
 		/* point or multipoint, use actual number of point count */
 		buffer[0] = c_int(CMD_MOVE_TO, c);
-	} else {
+	}
+	else
+	{
 		/* line or polygon, assume count 1 */
 		buffer[0] = c_int(CMD_MOVE_TO, 1);
 		/* line command with move point subtracted from count */
@@ -149,7 +163,7 @@ static uint32_t encode_ptarray(struct mvt_agg_context *ctx, enum mvt_type type,
 	return offset;
 }
 
-static uint32_t encode_ptarray_initial(struct mvt_agg_context *ctx,
+static uint32_t encode_ptarray_initial(mvt_agg_context *ctx,
 				       enum mvt_type type,
 				       POINTARRAY *pa, uint32_t *buffer)
 {
@@ -157,7 +171,7 @@ static uint32_t encode_ptarray_initial(struct mvt_agg_context *ctx,
 	return encode_ptarray(ctx, type, pa, buffer, &px, &py);
 }
 
-static void encode_point(struct mvt_agg_context *ctx, LWPOINT *point)
+static void encode_point(mvt_agg_context *ctx, LWPOINT *point)
 {
 	VectorTile__Tile__Feature *feature = ctx->feature;
 	feature->type = VECTOR_TILE__TILE__GEOM_TYPE__POINT;
@@ -167,7 +181,7 @@ static void encode_point(struct mvt_agg_context *ctx, LWPOINT *point)
 	encode_ptarray_initial(ctx, MVT_POINT, point->point, feature->geometry);
 }
 
-static void encode_mpoint(struct mvt_agg_context *ctx, LWMPOINT *mpoint)
+static void encode_mpoint(mvt_agg_context *ctx, LWMPOINT *mpoint)
 {
 	size_t c;
 	VectorTile__Tile__Feature *feature = ctx->feature;
@@ -181,7 +195,7 @@ static void encode_mpoint(struct mvt_agg_context *ctx, LWMPOINT *mpoint)
 		lwline->points, feature->geometry);
 }
 
-static void encode_line(struct mvt_agg_context *ctx, LWLINE *lwline)
+static void encode_line(mvt_agg_context *ctx, LWLINE *lwline)
 {
 	size_t c;
 	VectorTile__Tile__Feature *feature = ctx->feature;
@@ -193,7 +207,7 @@ static void encode_line(struct mvt_agg_context *ctx, LWLINE *lwline)
 		lwline->points, feature->geometry);
 }
 
-static void encode_mline(struct mvt_agg_context *ctx, LWMLINE *lwmline)
+static void encode_mline(mvt_agg_context *ctx, LWMLINE *lwmline)
 {
 	uint32_t i;
 	int32_t px = 0, py = 0;
@@ -211,7 +225,7 @@ static void encode_mline(struct mvt_agg_context *ctx, LWMLINE *lwmline)
 	feature->n_geometry = offset;
 }
 
-static void encode_poly(struct mvt_agg_context *ctx, LWPOLY *lwpoly)
+static void encode_poly(mvt_agg_context *ctx, LWPOLY *lwpoly)
 {
 	uint32_t i;
 	int32_t px = 0, py = 0;
@@ -229,7 +243,7 @@ static void encode_poly(struct mvt_agg_context *ctx, LWPOLY *lwpoly)
 	feature->n_geometry = offset;
 }
 
-static void encode_mpoly(struct mvt_agg_context *ctx, LWMPOLY *lwmpoly)
+static void encode_mpoly(mvt_agg_context *ctx, LWMPOLY *lwmpoly)
 {
 	uint32_t i, j;
 	int32_t px = 0, py = 0;
@@ -250,11 +264,12 @@ static void encode_mpoly(struct mvt_agg_context *ctx, LWMPOLY *lwmpoly)
 	feature->n_geometry = offset;
 }
 
-static void encode_geometry(struct mvt_agg_context *ctx, LWGEOM *lwgeom)
+static void encode_geometry(mvt_agg_context *ctx, LWGEOM *lwgeom)
 {
 	int type = lwgeom->type;
 
-	switch (type) {
+	switch (type)
+	{
 	case POINTTYPE:
 		return encode_point(ctx, (LWPOINT*)lwgeom);
 	case LINETYPE:
@@ -272,7 +287,7 @@ static void encode_geometry(struct mvt_agg_context *ctx, LWGEOM *lwgeom)
 	}
 }
 
-static TupleDesc get_tuple_desc(struct mvt_agg_context *ctx)
+static TupleDesc get_tuple_desc(mvt_agg_context *ctx)
 {
 	Oid tupType = HeapTupleHeaderGetTypeId(ctx->row);
 	int32 tupTypmod = HeapTupleHeaderGetTypMod(ctx->row);
@@ -280,17 +295,16 @@ static TupleDesc get_tuple_desc(struct mvt_agg_context *ctx)
 	return tupdesc;
 }
 
-static uint32_t get_key_index(struct mvt_agg_context *ctx, char *name)
+static uint32_t get_key_index_with_size(mvt_agg_context *ctx, const char *name, size_t size)
 {
 	struct mvt_kv_key *kv;
-	size_t size = strlen(name);
 	HASH_FIND(hh, ctx->keys_hash, name, size, kv);
 	if (!kv)
-		return -1;
+		return UINT32_MAX;
 	return kv->id;
 }
 
-static uint32_t add_key(struct mvt_agg_context *ctx, char *name)
+static uint32_t add_key(mvt_agg_context *ctx, char *name)
 {
 	struct mvt_kv_key *kv;
 	size_t size = strlen(name);
@@ -301,49 +315,67 @@ static uint32_t add_key(struct mvt_agg_context *ctx, char *name)
 	return kv->id;
 }
 
-static void parse_column_keys(struct mvt_agg_context *ctx)
+static void parse_column_keys(mvt_agg_context *ctx)
 {
-	TupleDesc tupdesc = get_tuple_desc(ctx);
-	int natts = tupdesc->natts;
-	uint32_t i;
+	uint32_t i, natts;
 	bool geom_found = false;
-	char *key;
+
 	POSTGIS_DEBUG(2, "parse_column_keys called");
 
-	for (i = 0; i < natts; i++) {
+	ctx->column_cache.tupdesc = get_tuple_desc(ctx);
+	natts = ctx->column_cache.tupdesc->natts;
+
+	ctx->column_cache.column_keys_index = palloc(sizeof(uint32_t) * natts);
+	ctx->column_cache.column_oid = palloc(sizeof(uint32_t) * natts);
+	ctx->column_cache.values = palloc(sizeof(Datum) * natts);
+	ctx->column_cache.nulls = palloc(sizeof(bool) * natts);
+
+	for (i = 0; i < natts; i++)
+	{
 #if POSTGIS_PGSQL_VERSION < 110
-		Oid typoid = getBaseType(tupdesc->attrs[i]->atttypid);
-		char *tkey = tupdesc->attrs[i]->attname.data;
+		Oid typoid = getBaseType(ctx->column_cache.tupdesc->attrs[i]->atttypid);
+		char *tkey = ctx->column_cache.tupdesc->attrs[i]->attname.data;
 #else
-		Oid typoid = getBaseType(tupdesc->attrs[i].atttypid);
-		char *tkey = tupdesc->attrs[i].attname.data;
+		Oid typoid = getBaseType(ctx->column_cache.tupdesc->attrs[i].atttypid);
+		char *tkey = ctx->column_cache.tupdesc->attrs[i].attname.data;
 #endif
+
+		ctx->column_cache.column_oid[i] = typoid;
 #if POSTGIS_PGSQL_VERSION >= 94
 		if (typoid == JSONBOID)
+		{
+			ctx->column_cache.column_keys_index[i] = UINT32_MAX;
 			continue;
+		}
 #endif
-		key = pstrdup(tkey);
-		if (ctx->geom_name == NULL) {
-			if (!geom_found && typoid == TypenameGetTypid("geometry")) {
-				ctx->geom_index = i;
-				geom_found = true;
-				continue;
-			}
-		} else {
-			if (!geom_found && strcmp(key, ctx->geom_name) == 0) {
+
+		if (ctx->geom_name == NULL)
+		{
+			if (!geom_found && typoid == TypenameGetTypid("geometry"))
+			{
 				ctx->geom_index = i;
 				geom_found = true;
 				continue;
 			}
 		}
-		add_key(ctx, key);
+		else
+		{
+			if (!geom_found && strcmp(tkey, ctx->geom_name) == 0)
+			{
+				ctx->geom_index = i;
+				geom_found = true;
+				continue;
+			}
+		}
+
+		ctx->column_cache.column_keys_index[i] = add_key(ctx, pstrdup(tkey));
 	}
+
 	if (!geom_found)
 		elog(ERROR, "parse_column_keys: no geometry column found");
-	ReleaseTupleDesc(tupdesc);
 }
 
-static void encode_keys(struct mvt_agg_context *ctx)
+static void encode_keys(mvt_agg_context *ctx)
 {
 	struct mvt_kv_key *kv;
 	size_t n_keys = ctx->keys_hash_i;
@@ -366,22 +398,28 @@ static VectorTile__Tile__Value *create_value()
 #define MVT_CREATE_VALUES(kvtype, hash, hasfield, valuefield) \
 { \
 	POSTGIS_DEBUG(2, "MVT_CREATE_VALUES called"); \
-	struct kvtype *kv; \
-	for (kv = ctx->hash; kv != NULL; kv=kv->hh.next) { \
-		VectorTile__Tile__Value *value = create_value(); \
-		value->hasfield = 1; \
-		value->valuefield = kv->valuefield; \
-		values[kv->id] = value; \
+	{ \
+		struct kvtype *kv; \
+		for (kv = ctx->hash; kv != NULL; kv=kv->hh.next) \
+		{ \
+			VectorTile__Tile__Value *value = create_value(); \
+			value->hasfield = 1; \
+			value->valuefield = kv->valuefield; \
+			values[kv->id] = value; \
+		} \
 	} \
 }
 
-static void encode_values(struct mvt_agg_context *ctx)
+static void encode_values(mvt_agg_context *ctx)
 {
-	POSTGIS_DEBUG(2, "encode_values called");
 	VectorTile__Tile__Value **values;
-	values = palloc(ctx->values_hash_i * sizeof(*values));
 	struct mvt_kv_string_value *kv;
-	for (kv = ctx->string_values_hash; kv != NULL; kv=kv->hh.next) {
+
+	POSTGIS_DEBUG(2, "encode_values called");
+
+	values = palloc(ctx->values_hash_i * sizeof(*values));
+	for (kv = ctx->string_values_hash; kv != NULL; kv=kv->hh.next)
+	{
 		VectorTile__Tile__Value *value = create_value();
 		value->string_value = kv->string_value;
 		values[kv->id] = value;
@@ -407,34 +445,48 @@ static void encode_values(struct mvt_agg_context *ctx)
 	HASH_CLEAR(hh, ctx->uint_values_hash);
 	HASH_CLEAR(hh, ctx->sint_values_hash);
 	HASH_CLEAR(hh, ctx->bool_values_hash);
+
+	pfree(ctx->column_cache.column_keys_index);
+	pfree(ctx->column_cache.column_oid);
+	pfree(ctx->column_cache.values);
+	pfree(ctx->column_cache.nulls);
+	ReleaseTupleDesc(ctx->column_cache.tupdesc);
+	memset(&ctx->column_cache, 0, sizeof(ctx->column_cache));
+
 }
 
 #define MVT_PARSE_VALUE(value, kvtype, hash, valuefield, size) \
 { \
 	POSTGIS_DEBUG(2, "MVT_PARSE_VALUE called"); \
-	struct kvtype *kv; \
-	HASH_FIND(hh, ctx->hash, &value, size, kv); \
-	if (!kv) { \
-		POSTGIS_DEBUG(4, "MVT_PARSE_VALUE value not found"); \
-		kv = palloc(sizeof(*kv)); \
-		POSTGIS_DEBUGF(4, "MVT_PARSE_VALUE new hash key: %d", \
-			ctx->values_hash_i); \
-		kv->id = ctx->values_hash_i++; \
-		kv->valuefield = value; \
-		HASH_ADD(hh, ctx->hash, valuefield, size, kv); \
+	{ \
+		struct kvtype *kv; \
+		HASH_FIND(hh, ctx->hash, &value, size, kv); \
+		if (!kv) \
+		{ \
+			POSTGIS_DEBUG(4, "MVT_PARSE_VALUE value not found"); \
+			kv = palloc(sizeof(*kv)); \
+			POSTGIS_DEBUGF(4, "MVT_PARSE_VALUE new hash key: %d", \
+				ctx->values_hash_i); \
+			kv->id = ctx->values_hash_i++; \
+			kv->valuefield = value; \
+			HASH_ADD(hh, ctx->hash, valuefield, size, kv); \
+		} \
+		tags[ctx->row_columns*2] = k; \
+		tags[ctx->row_columns*2+1] = kv->id; \
 	} \
-	tags[ctx->c*2] = k; \
-	tags[ctx->c*2+1] = kv->id; \
 }
 
 #define MVT_PARSE_INT_VALUE(value) \
 { \
-	if (value >= 0) { \
+	if (value >= 0) \
+	{ \
 		uint64_t cvalue = value; \
 		MVT_PARSE_VALUE(cvalue, mvt_kv_uint_value, \
 				uint_values_hash, uint_value, \
 				sizeof(uint64_t)) \
-	} else { \
+	} \
+	else \
+	{ \
 		int64_t cvalue = value; \
 		MVT_PARSE_VALUE(cvalue, mvt_kv_sint_value, \
 				sint_values_hash, sint_value, \
@@ -454,14 +506,14 @@ static void encode_values(struct mvt_agg_context *ctx)
 	MVT_PARSE_INT_VALUE(value); \
 }
 
-static void add_value_as_string(struct mvt_agg_context *ctx,
-	char *value, uint32_t *tags, uint32_t k)
+static void add_value_as_string_with_size(mvt_agg_context *ctx,
+	char *value, size_t size, uint32_t *tags, uint32_t k)
 {
 	struct mvt_kv_string_value *kv;
-	size_t size = strlen(value);
 	POSTGIS_DEBUG(2, "add_value_as_string called");
 	HASH_FIND(hh, ctx->string_values_hash, value, size, kv);
-	if (!kv) {
+	if (!kv)
+	{
 		POSTGIS_DEBUG(4, "add_value_as_string value not found");
 		kv = palloc(sizeof(*kv));
 		POSTGIS_DEBUGF(4, "add_value_as_string new hash key: %d",
@@ -471,11 +523,17 @@ static void add_value_as_string(struct mvt_agg_context *ctx,
 		HASH_ADD_KEYPTR(hh, ctx->string_values_hash, kv->string_value,
 			size, kv);
 	}
-	tags[ctx->c*2] = k;
-	tags[ctx->c*2+1] = kv->id;
+	tags[ctx->row_columns*2] = k;
+	tags[ctx->row_columns*2+1] = kv->id;
 }
 
-static void parse_datum_as_string(struct mvt_agg_context *ctx, Oid typoid,
+static void add_value_as_string(mvt_agg_context *ctx,
+	char *value, uint32_t *tags, uint32_t k)
+{
+	return add_value_as_string_with_size(ctx, value, strlen(value), tags, k);
+}
+
+static void parse_datum_as_string(mvt_agg_context *ctx, Oid typoid,
 	Datum datum, uint32_t *tags, uint32_t k)
 {
 	Oid foutoid;
@@ -489,7 +547,7 @@ static void parse_datum_as_string(struct mvt_agg_context *ctx, Oid typoid,
 }
 
 #if POSTGIS_PGSQL_VERSION >= 94
-static uint32_t *parse_jsonb(struct mvt_agg_context *ctx, Jsonb *jb,
+static uint32_t *parse_jsonb(mvt_agg_context *ctx, Jsonb *jb,
 	uint32_t *tags)
 {
 	JsonbIterator *it;
@@ -503,48 +561,63 @@ static uint32_t *parse_jsonb(struct mvt_agg_context *ctx, Jsonb *jb,
 
 	it = JsonbIteratorInit(&jb->root);
 
-	while ((r = JsonbIteratorNext(&it, &v, skipNested)) != WJB_DONE) {
+	while ((r = JsonbIteratorNext(&it, &v, skipNested)) != WJB_DONE)
+	{
 		skipNested = true;
 
-		if (r == WJB_KEY && v.type != jbvNull) {
-			char *key;
-			key = palloc(v.val.string.len + 1 * sizeof(char));
-			memcpy(key, v.val.string.val, v.val.string.len);
-			key[v.val.string.len] = '\0';
+		if (r == WJB_KEY && v.type != jbvNull)
+		{
 
-			k = get_key_index(ctx, key);
-			if (k == -1) {
+			k = get_key_index_with_size(ctx, v.val.string.val, v.val.string.len);
+			if (k == UINT32_MAX)
+			{
+				char *key;
 				uint32_t newSize = ctx->keys_hash_i + 1;
+
+				key = palloc(v.val.string.len + 1);
+				memcpy(key, v.val.string.val, v.val.string.len);
+				key[v.val.string.len] = '\0';
+
 				tags = repalloc(tags, newSize * 2 * sizeof(*tags));
 				k = add_key(ctx, key);
 			}
 
 			r = JsonbIteratorNext(&it, &v, skipNested);
 
-			if (v.type == jbvString) {
+			if (v.type == jbvString)
+			{
 				char *value;
-				value = palloc(v.val.string.len + 1 * sizeof(char));
+				value = palloc(v.val.string.len + 1);
 				memcpy(value, v.val.string.val, v.val.string.len);
 				value[v.val.string.len] = '\0';
 				add_value_as_string(ctx, value, tags, k);
-				ctx->c++;
-			} else if (v.type == jbvBool) {
+				ctx->row_columns++;
+			}
+			else if (v.type == jbvBool)
+			{
 				MVT_PARSE_VALUE(v.val.boolean, mvt_kv_bool_value,
 					bool_values_hash, bool_value, sizeof(protobuf_c_boolean));
-				ctx->c++;
-			} else if (v.type == jbvNumeric) {
+				ctx->row_columns++;
+			}
+			else if (v.type == jbvNumeric)
+			{
 				char *str;
+				double d;
+				long l;
 				str = DatumGetCString(DirectFunctionCall1(numeric_out,
 					PointerGetDatum(v.val.numeric)));
-				double d = strtod(str, NULL);
-				long l = strtol(str, NULL, 10);
-				if ((long) d != l) {
+				d = strtod(str, NULL);
+				l = strtol(str, NULL, 10);
+				if ((long) d != l)
+				{
 					MVT_PARSE_VALUE(d, mvt_kv_double_value, double_values_hash,
 						double_value, sizeof(double));
-				} else {
+				}
+				else
+				{
 					MVT_PARSE_INT_VALUE(l);
 				}
-				ctx->c++;
+				ctx->row_columns++;
 			}
 		}
 	}
@@ -553,53 +626,69 @@ static uint32_t *parse_jsonb(struct mvt_agg_context *ctx, Jsonb *jb,
 }
 #endif
 
-static void parse_values(struct mvt_agg_context *ctx)
+static void parse_values(mvt_agg_context *ctx)
 {
-	POSTGIS_DEBUG(2, "parse_values called");
 	uint32_t n_keys = ctx->keys_hash_i;
 	uint32_t *tags = palloc(n_keys * 2 * sizeof(*tags));
-	bool isnull;
-	uint32_t i, k;
-	TupleDesc tupdesc = get_tuple_desc(ctx);
-	int natts = tupdesc->natts;
-	ctx->c = 0;
+	uint32_t i;
+	mvt_column_cache cc = ctx->column_cache;
+	uint32_t natts = (uint32_t) cc.tupdesc->natts;
+
+	HeapTupleData tuple;
+
+	POSTGIS_DEBUG(2, "parse_values called");
+	ctx->row_columns = 0;
+
+	/* Build a temporary HeapTuple control structure */
+	tuple.t_len = HeapTupleHeaderGetDatumLength(ctx->row);
+	ItemPointerSetInvalid(&(tuple.t_self));
+	tuple.t_tableOid = InvalidOid;
+	tuple.t_data = ctx->row;
+
+	/* We use heap_deform_tuple as it costs only O(N) vs O(N^2) of GetAttributeByNum */
+	heap_deform_tuple(&tuple, cc.tupdesc, cc.values, cc.nulls);
 
 	POSTGIS_DEBUGF(3, "parse_values natts: %d", natts);
 
-	for (i = 0; i < natts; i++) {
+	for (i = 0; i < natts; i++)
+	{
 		char *key;
 		Oid typoid;
-		Datum datum;
+		uint32_t k;
+		Datum datum = cc.values[i];
 
 		if (i == ctx->geom_index)
 			continue;
 
-#if POSTGIS_PGSQL_VERSION < 110
-		key = tupdesc->attrs[i]->attname.data;
-		typoid = getBaseType(tupdesc->attrs[i]->atttypid);
-#else
-		key = tupdesc->attrs[i].attname.data;
-		typoid = getBaseType(tupdesc->attrs[i].atttypid);
-#endif
-		datum = GetAttributeByNum(ctx->row, i+1, &isnull);
-		k = get_key_index(ctx, key);
-		if (isnull) {
+		if (cc.nulls[i])
+		{
 			POSTGIS_DEBUG(3, "parse_values isnull detected");
 			continue;
 		}
+
+#if POSTGIS_PGSQL_VERSION < 110
+		key = cc.tupdesc->attrs[i]->attname.data;
+#else
+		key = cc.tupdesc->attrs[i].attname.data;
+#endif
+		k = cc.column_keys_index[i];
+		typoid = cc.column_oid[i];
+
 #if POSTGIS_PGSQL_VERSION >= 94
-		if (k == -1 && typoid != JSONBOID)
+		if (k == UINT32_MAX && typoid != JSONBOID)
 			elog(ERROR, "parse_values: unexpectedly could not find parsed key name '%s'", key);
-		if (typoid == JSONBOID) {
+		if (typoid == JSONBOID)
+		{
 			tags = parse_jsonb(ctx, DatumGetJsonbP(datum), tags);
 			continue;
 		}
 #else
-		if (k == -1)
+		if (k == UINT32_MAX)
 			elog(ERROR, "parse_values: unexpectedly could not find parsed key name '%s'", key);
 #endif
 
-		switch (typoid) {
+		switch (typoid)
+		{
 		case BOOLOID:
 			MVT_PARSE_DATUM(protobuf_c_boolean, mvt_kv_bool_value,
 				bool_values_hash, bool_value,
@@ -628,27 +717,46 @@ static void parse_values(struct mvt_agg_context *ctx)
 			parse_datum_as_string(ctx, typoid, datum, tags, k);
 			break;
 		}
-		ctx->c++;
+		ctx->row_columns++;
 	}
 
-	ReleaseTupleDesc(tupdesc);
 
-	ctx->feature->n_tags = ctx->c * 2;
+	ctx->feature->n_tags = ctx->row_columns * 2;
 	ctx->feature->tags = tags;
 
 	POSTGIS_DEBUGF(3, "parse_values n_tags %zd", ctx->feature->n_tags);
 }
-static int max_type(LWCOLLECTION *lwcoll)
+
+/**
+ * In place process a collection to find a concrete geometry
+ * object and expose that as the actual object. Will some
+ * geom be lost? Sure, but your MVT renderer couldn't
+ * draw it anyways.
+ */
+static void
+lwgeom_to_basic_type(LWGEOM *geom)
 {
-	int i, max = POINTTYPE;
-	for (i = 0; i < lwcoll->ngeoms; i++) {
-		uint8_t type = lwcoll->geoms[i]->type;
-		if (type == POLYGONTYPE || type == MULTIPOLYGONTYPE)
-			return POLYGONTYPE;
-		else if (type == LINETYPE || type == MULTILINETYPE)
-			max = LINETYPE;
+	if (lwgeom_get_type(geom) == COLLECTIONTYPE)
+	{
+		/* MVT doesn't handle generic collections, so we */
+		/* need to strip them down to a typed collection */
+		/* by finding the largest basic type available and */
+		/* using that as the basis of a typed collection. */
+		LWCOLLECTION *g = (LWCOLLECTION*)geom;
+		LWCOLLECTION *gc;
+		uint32_t i, maxtype = 0;
+		for (i = 0; i < g->ngeoms; i++)
+		{
+			LWGEOM *sg = g->geoms[i];
+			if (sg->type > maxtype && sg->type < COLLECTIONTYPE)
+				maxtype = sg->type;
+		}
+		if (maxtype > 3) maxtype -= 3;
+		/* Force the working geometry to be a simpler version */
+		/* of itself */
+		gc = lwcollection_extract(g, maxtype);
+		*g = *gc;
 	}
-	return max;
 }
 
 /**
@@ -656,62 +764,71 @@ static int max_type(LWCOLLECTION *lwcoll)
  *
  * Makes best effort to keep validity. Might collapse geometry into lower
  * dimension.
+ *
+ * NOTE: modifies in place if possible (not currently possible for polygons)
  */
-LWGEOM *mvt_geom(const LWGEOM *lwgeom, const GBOX *gbox, uint32_t extent, uint32_t buffer,
+LWGEOM *mvt_geom(LWGEOM *lwgeom, const GBOX *gbox, uint32_t extent, uint32_t buffer,
 	bool clip_geom)
 {
 	AFFINE affine;
 	gridspec grid;
-	LWGEOM *lwgeom_out = NULL;
 	double width = gbox->xmax - gbox->xmin;
 	double height = gbox->ymax - gbox->ymin;
-	double resx = width / extent;
-	double fx = extent / width;
-	double fy = -(extent / height);
-	double buffer_map_xunits = resx * buffer;
-	const GBOX *ggbox;
+	double resx, resy, res, fx, fy;
+	int preserve_collapsed = LW_TRUE;
 	POSTGIS_DEBUG(2, "mvt_geom called");
 
 	/* Short circuit out on EMPTY */
 	if (lwgeom_is_empty(lwgeom))
 		return NULL;
 
-	ggbox = lwgeom_get_bbox(lwgeom);
 	if (width == 0 || height == 0)
 		elog(ERROR, "mvt_geom: bounds width or height cannot be 0");
 
 	if (extent == 0)
 		elog(ERROR, "mvt_geom: extent cannot be 0");
 
-	if (clip_geom) {
-		GBOX *bgbox = gbox_copy(gbox);
-		gbox_expand(bgbox, buffer_map_xunits);
-		if (!gbox_overlaps_2d(ggbox, bgbox)) {
+	resx = width / extent;
+	resy = height / extent;
+	res = (resx < resy ? resx : resy)/2;
+	fx = extent / width;
+	fy = -(extent / height);
+
+	/* Remove all non-essential points (under the output resolution) */
+	lwgeom_remove_repeated_points_in_place(lwgeom, res);
+	lwgeom_simplify_in_place(lwgeom, res, preserve_collapsed);
+
+	/* If geometry has disappeared, you're done */
+	if (lwgeom_is_empty(lwgeom))
+		return NULL;
+
+	if (clip_geom)
+	{
+		// We need to add an extra half pixel to include the points that
+		// fall into the bbox only after the coordinate transformation
+		double buffer_map_xunits = nextafterf(res, 0.0) + resx * buffer;
+		GBOX bgbox;
+		const GBOX *lwgeom_gbox = lwgeom_get_bbox(lwgeom);
+		bgbox = *gbox;
+		gbox_expand(&bgbox, buffer_map_xunits);
+		if (!gbox_overlaps_2d(lwgeom_gbox, &bgbox))
+		{
 			POSTGIS_DEBUG(3, "mvt_geom: geometry outside clip box");
 			return NULL;
 		}
-		if (!gbox_contains_2d(bgbox, ggbox)) {
-			double x0 = bgbox->xmin;
-			double y0 = bgbox->ymin;
-			double x1 = bgbox->xmax;
-			double y1 = bgbox->ymax;
-#if POSTGIS_GEOS_VERSION < 35
-			LWPOLY *lwenv = lwpoly_construct_envelope(0, x0, y0, x1, y1);
-			lwgeom_out = lwgeom_intersection(lwgeom, lwpoly_as_lwgeom(lwenv));
-			lwpoly_free(lwenv);
-#else
-			lwgeom_out = lwgeom_clip_by_rect(lwgeom, x0, y0, x1, y1);
-#endif
+		if (!gbox_contains_2d(&bgbox, lwgeom_gbox))
+		{
+			double x0 = bgbox.xmin;
+			double y0 = bgbox.ymin;
+			double x1 = bgbox.xmax;
+			double y1 = bgbox.ymax;
+			lwgeom = lwgeom_clip_by_rect(lwgeom, x0, y0, x1, y1);
 			POSTGIS_DEBUG(3, "mvt_geom: no geometry after clip");
-			if (lwgeom_out == NULL || lwgeom_is_empty(lwgeom_out))
+			if (lwgeom == NULL || lwgeom_is_empty(lwgeom))
 				return NULL;
 		}
 	}
 
-	/* if no clip output deep clone original to avoid mutation */
-	if (lwgeom_out == NULL)
-		lwgeom_out = lwgeom_clone_deep(lwgeom);
-
 	/* transform to tile coordinate space */
 	memset(&affine, 0, sizeof(affine));
 	affine.afac = fx;
@@ -719,7 +836,7 @@ LWGEOM *mvt_geom(const LWGEOM *lwgeom, const GBOX *gbox, uint32_t extent, uint32
 	affine.ifac = 1;
 	affine.xoff = -gbox->xmin * fx;
 	affine.yoff = -gbox->ymax * fy;
-	lwgeom_affine(lwgeom_out, &affine);
+	lwgeom_affine(lwgeom, &affine);
 
 	/* snap to integer precision, removing duplicate points */
 	memset(&grid, 0, sizeof(gridspec));
@@ -727,42 +844,47 @@ LWGEOM *mvt_geom(const LWGEOM *lwgeom, const GBOX *gbox, uint32_t extent, uint32
 	grid.ipy = 0;
 	grid.xsize = 1;
 	grid.ysize = 1;
-	lwgeom_out = lwgeom_grid(lwgeom_out, &grid);
+	lwgeom_grid_in_place(lwgeom, &grid);
 
-	if (lwgeom_out == NULL || lwgeom_is_empty(lwgeom_out))
+	if (lwgeom == NULL || lwgeom_is_empty(lwgeom))
 		return NULL;
 
-	/* if polygon(s) make valid and force clockwise as per MVT spec */
-	if (lwgeom_out->type == POLYGONTYPE ||
-		lwgeom_out->type == MULTIPOLYGONTYPE) {
-		lwgeom_out = lwgeom_make_valid(lwgeom_out);
-		lwgeom_force_clockwise(lwgeom_out);
+
+	if (lwgeom->type == POLYGONTYPE ||
+		lwgeom->type == MULTIPOLYGONTYPE ||
+		lwgeom->type == COLLECTIONTYPE)
+	{
+		/* Force validation as per MVT spec */
+		lwgeom = lwgeom_make_valid(lwgeom);
+
+		/* Drop type changes tp play nice with MVT renderers */
+		if (!(lwgeom->type == POLYGONTYPE ||
+			lwgeom->type == MULTIPOLYGONTYPE ||
+			lwgeom->type == COLLECTIONTYPE))
+		{
+			lwgeom_free(lwgeom);
+			return NULL;
+		}
+
+		/* In image coordinates CW actually comes out a CCW, so we reverse */
+		lwgeom_force_clockwise(lwgeom);
+		lwgeom_reverse_in_place(lwgeom);
 	}
 
 	/* if geometry collection extract highest dimensional geometry type */
-	if (lwgeom_out->type == COLLECTIONTYPE) {
-		LWCOLLECTION *lwcoll = lwgeom_as_lwcollection(lwgeom_out);
-		lwgeom_out = lwcollection_as_lwgeom(
-			lwcollection_extract(lwcoll, max_type(lwcoll)));
-		lwgeom_out = lwgeom_homogenize(lwgeom_out);
-		/* if polygon(s) make valid and force clockwise as per MVT spec */
-		if (lwgeom_out->type == POLYGONTYPE ||
-			lwgeom_out->type == MULTIPOLYGONTYPE) {
-			lwgeom_out = lwgeom_make_valid(lwgeom_out);
-			lwgeom_force_clockwise(lwgeom_out);
-		}
-	}
+	if (lwgeom->type == COLLECTIONTYPE)
+		lwgeom_to_basic_type(lwgeom);
 
-	if (lwgeom_out == NULL || lwgeom_is_empty(lwgeom_out))
+	if (lwgeom == NULL || lwgeom_is_empty(lwgeom))
 		return NULL;
 
-	return lwgeom_out;
+	return lwgeom;
 }
 
 /**
  * Initialize aggregation context.
  */
-void mvt_agg_init_context(struct mvt_agg_context *ctx)
+void mvt_agg_init_context(mvt_agg_context *ctx)
 {
 	VectorTile__Tile__Layer *layer;
 
@@ -771,6 +893,7 @@ void mvt_agg_init_context(struct mvt_agg_context *ctx)
 	if (ctx->extent == 0)
 		elog(ERROR, "mvt_agg_init_context: extent cannot be 0");
 
+	ctx->tile = NULL;
 	ctx->features_capacity = FEATURES_CAPACITY_INITIAL;
 	ctx->keys_hash = NULL;
 	ctx->string_values_hash = NULL;
@@ -781,6 +904,9 @@ void mvt_agg_init_context(struct mvt_agg_context *ctx)
 	ctx->bool_values_hash = NULL;
 	ctx->values_hash_i = 0;
 	ctx->keys_hash_i = 0;
+	ctx->geom_index = UINT32_MAX;
+
+	memset(&ctx->column_cache, 0, sizeof(ctx->column_cache));
 
 	layer = palloc(sizeof(*layer));
 	vector_tile__tile__layer__init(layer);
@@ -801,7 +927,7 @@ void mvt_agg_init_context(struct mvt_agg_context *ctx)
  * Allocates a new feature, increment feature counter and
  * encode geometry and properties into it.
  */
-void mvt_agg_transfn(struct mvt_agg_context *ctx)
+void mvt_agg_transfn(mvt_agg_context *ctx)
 {
 	bool isnull = false;
 	Datum datum;
@@ -809,9 +935,11 @@ void mvt_agg_transfn(struct mvt_agg_context *ctx)
 	LWGEOM *lwgeom;
 	VectorTile__Tile__Feature *feature;
 	VectorTile__Tile__Layer *layer = ctx->layer;
+/*	VectorTile__Tile__Feature **features = layer->features; */
 	POSTGIS_DEBUG(2, "mvt_agg_transfn called");
 
-	if (layer->n_features >= ctx->features_capacity) {
+	if (layer->n_features >= ctx->features_capacity)
+	{
 		size_t new_capacity = ctx->features_capacity * 2;
 		layer->features = repalloc(layer->features, new_capacity *
 			sizeof(*layer->features));
@@ -819,7 +947,7 @@ void mvt_agg_transfn(struct mvt_agg_context *ctx)
 		POSTGIS_DEBUGF(3, "mvt_agg_transfn new_capacity: %zd", new_capacity);
 	}
 
-	if (layer->n_features == 0)
+	if (ctx->geom_index == UINT32_MAX)
 		parse_column_keys(ctx);
 
 	datum = GetAttributeByNum(ctx->row, ctx->geom_index + 1, &isnull);
@@ -849,45 +977,264 @@ void mvt_agg_transfn(struct mvt_agg_context *ctx)
 	parse_values(ctx);
 }
 
+static VectorTile__Tile * mvt_ctx_to_tile(mvt_agg_context *ctx)
+{
+	int n_layers = 1;
+	VectorTile__Tile *tile;
+	encode_keys(ctx);
+	encode_values(ctx);
+
+	tile = palloc(sizeof(VectorTile__Tile));
+	vector_tile__tile__init(tile);
+	tile->layers = palloc(sizeof(VectorTile__Tile__Layer*) * n_layers);
+	tile->layers[0] = ctx->layer;
+	tile->n_layers = n_layers;
+	return tile;
+}
+
+static bytea *mvt_ctx_to_bytea(mvt_agg_context *ctx)
+{
+	/* Fill out the file slot, if it's not already filled. */
+	/* We should only have a filled slow when all the work of building */
+	/* out the data is complete, so after a serialize/deserialize cycle */
+	/* or after a context combine */
+	size_t len;
+	bytea *ba;
+
+	if (!ctx->tile)
+	{
+		ctx->tile = mvt_ctx_to_tile(ctx);
+	}
+
+	/* Zero features => empty bytea output */
+	if (ctx && ctx->layer && ctx->layer->n_features == 0)
+	{
+		bytea *ba = palloc(VARHDRSZ);
+		SET_VARSIZE(ba, VARHDRSZ);
+		return ba;
+	}
+
+	/* Serialize the Tile */
+	len = VARHDRSZ + vector_tile__tile__get_packed_size(ctx->tile);
+	ba = palloc(len);
+	vector_tile__tile__pack(ctx->tile, (uint8_t*)VARDATA(ba));
+	SET_VARSIZE(ba, len);
+	return ba;
+}
+
+
+bytea * mvt_ctx_serialize(mvt_agg_context *ctx)
+{
+	return mvt_ctx_to_bytea(ctx);
+}
+
+static void * mvt_allocator(__attribute__((__unused__)) void *data, size_t size)
+{
+	return palloc(size);
+}
+
+static void mvt_deallocator(__attribute__((__unused__)) void *data, void *ptr)
+{
+	return pfree(ptr);
+}
+
+mvt_agg_context * mvt_ctx_deserialize(const bytea *ba)
+{
+	ProtobufCAllocator allocator =
+	{
+		mvt_allocator,
+		mvt_deallocator,
+		NULL
+	};
+
+	size_t len = VARSIZE(ba) - VARHDRSZ;
+	VectorTile__Tile *tile = vector_tile__tile__unpack(&allocator, len, (uint8_t*)VARDATA(ba));
+	mvt_agg_context *ctx = palloc(sizeof(mvt_agg_context));
+	memset(ctx, 0, sizeof(mvt_agg_context));
+	ctx->tile = tile;
+	return ctx;
+}
+
+static VectorTile__Tile__Value *
+tile_value_copy(const VectorTile__Tile__Value *value)
+{
+	VectorTile__Tile__Value *nvalue = palloc(sizeof(VectorTile__Tile__Value));
+	memcpy(nvalue, value, sizeof(VectorTile__Tile__Value));
+	if (value->string_value)
+		nvalue->string_value = pstrdup(value->string_value);
+	return nvalue;
+}
+
+static VectorTile__Tile__Feature *
+tile_feature_copy(const VectorTile__Tile__Feature *feature, int key_offset, int value_offset)
+{
+	uint32_t i;
+	VectorTile__Tile__Feature *nfeature;
+
+	/* Null in => Null out */
+	if (!feature) return NULL;
+
+	/* Init object */
+	nfeature = palloc(sizeof(VectorTile__Tile__Feature));
+	vector_tile__tile__feature__init(nfeature);
+
+	/* Copy settings straight over */
+	nfeature->has_id = feature->has_id;
+	nfeature->id = feature->id;
+	nfeature->has_type = feature->has_type;
+	nfeature->type = feature->type;
+
+	/* Copy tags over, offsetting indexes so they match the dictionaries */
+	/* at the Tile_Layer level */
+	if (feature->n_tags > 0)
+	{
+		nfeature->n_tags = feature->n_tags;
+		nfeature->tags = palloc(sizeof(uint32_t)*feature->n_tags);
+		for (i = 0; i < feature->n_tags/2; i++)
+		{
+			nfeature->tags[2*i] = feature->tags[2*i] + key_offset;
+			nfeature->tags[2*i+1] = feature->tags[2*i+1] + value_offset;
+		}
+	}
+
+	/* Copy the raw geometry data over literally */
+	if (feature->n_geometry > 0)
+	{
+		nfeature->n_geometry = feature->n_geometry;
+		nfeature->geometry = palloc(sizeof(uint32_t)*feature->n_geometry);
+		memcpy(nfeature->geometry, feature->geometry, sizeof(uint32_t)*feature->n_geometry);
+	}
+
+	/* Done */
+	return nfeature;
+}
+
+static VectorTile__Tile__Layer *
+vectortile_layer_combine(const VectorTile__Tile__Layer *layer1, const VectorTile__Tile__Layer *layer2)
+{
+	uint32_t i, j;
+	int key2_offset, value2_offset;
+	VectorTile__Tile__Layer *layer = palloc(sizeof(VectorTile__Tile__Layer));
+	vector_tile__tile__layer__init(layer);
+
+	/* Take globals from layer1 */
+	layer->version = layer1->version;
+	layer->name = pstrdup(layer1->name);
+	layer->has_extent = layer1->has_extent;
+	layer->extent = layer1->extent;
+
+	/* Copy keys into new layer */
+	j = 0;
+	layer->n_keys = layer1->n_keys + layer2->n_keys;
+	layer->keys = layer->n_keys ? palloc(layer->n_keys * sizeof(void*)) : NULL;
+	for (i = 0; i < layer1->n_keys; i++)
+		layer->keys[j++] = pstrdup(layer1->keys[i]);
+	key2_offset = j;
+	for (i = 0; i < layer2->n_keys; i++)
+		layer->keys[j++] = pstrdup(layer2->keys[i]);
+
+	/* Copy values into new layer */
+	/* TODO, apply hash logic here too, so that merged tiles */
+	/* retain unique value maps */
+	layer->n_values = layer1->n_values + layer2->n_values;
+	layer->values = layer->n_values ? palloc(layer->n_values * sizeof(void*)) : NULL;
+	j = 0;
+	for (i = 0; i < layer1->n_values; i++)
+		layer->values[j++] = tile_value_copy(layer1->values[i]);
+	value2_offset = j;
+	for (i = 0; i < layer2->n_values; i++)
+		layer->values[j++] = tile_value_copy(layer2->values[i]);
+
+
+	layer->n_features = layer1->n_features + layer2->n_features;
+	layer->features = layer->n_features ? palloc(layer->n_features * sizeof(void*)) : NULL;
+	j = 0;
+	for (i = 0; i < layer1->n_features; i++)
+		layer->features[j++] = tile_feature_copy(layer1->features[i], 0, 0);
+	for (i = 0; i < layer2->n_features; i++)
+		layer->features[j++] = tile_feature_copy(layer2->features[i], key2_offset, value2_offset);
+
+	return layer;
+}
+
+
+static VectorTile__Tile *
+vectortile_tile_combine(VectorTile__Tile *tile1, VectorTile__Tile *tile2)
+{
+	uint32_t i, j;
+	VectorTile__Tile *tile;
+
+	/* Hopelessly messing up memory ownership here */
+	if (tile1->n_layers == 0 && tile2->n_layers == 0)
+		return tile1;
+	else if (tile1->n_layers == 0)
+		return tile2;
+	else if (tile2->n_layers == 0)
+		return tile1;
+
+	tile = palloc(sizeof(VectorTile__Tile));
+	vector_tile__tile__init(tile);
+	tile->layers = palloc(sizeof(void*));
+	tile->n_layers = 0;
+
+	/* Merge all matching layers in the files (basically always only one) */
+	for (i = 0; i < tile1->n_layers; i++)
+	{
+		for (j = 0; j < tile2->n_layers; j++)
+		{
+			VectorTile__Tile__Layer *l1 = tile1->layers[i];
+			VectorTile__Tile__Layer *l2 = tile2->layers[j];
+			if (strcmp(l1->name, l2->name)==0)
+			{
+				VectorTile__Tile__Layer *layer = vectortile_layer_combine(l1, l2);
+				if (!layer)
+					continue;
+				tile->layers[tile->n_layers++] = layer;
+				/* Add a spare slot at the end of the array */
+				tile->layers = repalloc(tile->layers, (tile->n_layers+1) * sizeof(void*));
+			}
+		}
+	}
+	return tile;
+}
+
+mvt_agg_context * mvt_ctx_combine(mvt_agg_context *ctx1, mvt_agg_context *ctx2)
+{
+	if (ctx1 || ctx2)
+	{
+		if (ctx1 && ! ctx2) return ctx1;
+		if (ctx2 && ! ctx1) return ctx2;
+		if (ctx1 && ctx2 && ctx1->tile && ctx2->tile)
+		{
+			mvt_agg_context *ctxnew = palloc(sizeof(mvt_agg_context));
+			memset(ctxnew, 0, sizeof(mvt_agg_context));
+			ctxnew->tile = vectortile_tile_combine(ctx1->tile, ctx2->tile);
+			return ctxnew;
+		}
+		else
+		{
+			elog(DEBUG2, "ctx1->tile = %p", ctx1->tile);
+			elog(DEBUG2, "ctx2->tile = %p", ctx2->tile);
+			elog(ERROR, "%s: unable to combine contexts where tile attribute is null", __func__);
+			return NULL;
+		}
+	}
+	else
+	{
+		return NULL;
+	}
+}
+
 /**
  * Finalize aggregation.
  *
  * Encode keys and values and put the aggregated Layer message into
  * a Tile message and returns it packed as a bytea.
  */
-uint8_t *mvt_agg_finalfn(struct mvt_agg_context *ctx)
+bytea *mvt_agg_finalfn(mvt_agg_context *ctx)
 {
-	VectorTile__Tile__Layer *layers[1];
-	VectorTile__Tile tile = VECTOR_TILE__TILE__INIT;
-	size_t len;
-	uint8_t *buf;
-
-	POSTGIS_DEBUG(2, "mvt_agg_finalfn called");
-	POSTGIS_DEBUGF(2, "mvt_agg_finalfn n_features == %zd", ctx->layer->n_features);
-
-	/* Zero features => empty bytea output */
-	if (ctx->layer->n_features == 0)
-	{
-		buf = palloc(VARHDRSZ);
-		SET_VARSIZE(buf, VARHDRSZ);
-		return buf;
-	}
-
-	encode_keys(ctx);
-	encode_values(ctx);
-
-	layers[0] = ctx->layer;
-
-	tile.n_layers = 1;
-	tile.layers = layers;
-
-	len = vector_tile__tile__get_packed_size(&tile);
-	buf = palloc(sizeof(*buf) * (len + VARHDRSZ));
-	vector_tile__tile__pack(&tile, buf + VARHDRSZ);
-
-	SET_VARSIZE(buf, VARHDRSZ + len);
-
-	return buf;
+	return mvt_ctx_to_bytea(ctx);
 }
 
+
 #endif
