@@ -9,7 +9,7 @@
  * Copyright (C) 2010-2011 David Zwarg <dzwarg@azavea.com>
  * Copyright (C) 2009-2011 Pierre Racine <pierre.racine@sbf.ulaval.ca>
  * Copyright (C) 2009-2011 Mateusz Loskot <mateusz@loskot.net>
- * Copyright (C) 2008-2009 Sandro Santilli <strk@keybit.net>
+ * Copyright (C) 2008-2009 Sandro Santilli <strk@kbt.io>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -629,14 +629,24 @@ rt_raster_serialize(rt_raster raster) {
 				ptr += 1;
 				break;
 			}
-			case PT_16BSI:
+			case PT_16BSI: {
+				int16_t v = band->nodataval;
+				memcpy(ptr, &v, 2);
+				ptr += 2;
+				break;
+			}
 			case PT_16BUI: {
 				uint16_t v = band->nodataval;
 				memcpy(ptr, &v, 2);
 				ptr += 2;
 				break;
 			}
-			case PT_32BSI:
+			case PT_32BSI: {
+				int32_t v = band->nodataval;
+				memcpy(ptr, &v, 4);
+				ptr += 4;
+				break;
+			}
 			case PT_32BUI: {
 				uint32_t v = band->nodataval;
 				memcpy(ptr, &v, 4);
@@ -688,11 +698,9 @@ rt_raster_serialize(rt_raster raster) {
 #endif
 
 		/* Pad up to 8-bytes boundary */
-		while ((uintptr_t) ptr % 8) {
+		while ((ptr-ret) % 8) {
 			*ptr = 0;
 			++ptr;
-
-			RASTER_DEBUGF(3, "PAD at %d", (uintptr_t) ptr % 8);
 		}
 
 		/* Consistency checking (ptr is pixbytes-aligned) */
@@ -731,7 +739,7 @@ rt_raster_deserialize(void* serialized, int header_only) {
 	 * See note on SET_VARSIZE in rt_raster_serialize function above.
 	 */
 
-	/* Allocate memory for deserialized raster header */ 
+	/* Allocate memory for deserialized raster header */
 	RASTER_DEBUG(3, "rt_raster_deserialize: Allocating memory for deserialized raster header");
 	rast = (rt_raster) rtalloc(sizeof (struct rt_raster_t));
 	if (!rast) {
