@@ -29,7 +29,7 @@ static void test_gdal_configured() {
 }
 
 static void test_gdal_drivers() {
-	int i;
+	uint32_t i;
 	uint32_t size;
 	rt_gdaldriver drv = NULL;
 
@@ -471,6 +471,9 @@ static void test_raster_to_gdal() {
 
 	if (gdal) CPLFree(gdal);
 
+	gdal = rt_raster_to_gdal(raster, srs, "PCIDSK", NULL, &gdalSize);
+	CU_ASSERT(gdal == NULL);
+
 	cu_free_raster(raster);
 }
 
@@ -510,8 +513,8 @@ static void test_gdal_to_raster() {
 	CU_ASSERT(gddrv != NULL);
 	CU_ASSERT_EQUAL(destroy, 0);
 	CU_ASSERT(gdds != NULL);
-	CU_ASSERT_EQUAL(GDALGetRasterXSize(gdds), width);
-	CU_ASSERT_EQUAL(GDALGetRasterYSize(gdds), height);
+	CU_ASSERT_EQUAL((uint32_t)GDALGetRasterXSize(gdds), width);
+	CU_ASSERT_EQUAL((uint32_t)GDALGetRasterYSize(gdds), height);
 
 	rast = rt_raster_from_gdal_dataset(gdds);
 	CU_ASSERT(rast != NULL);
@@ -556,8 +559,8 @@ static void test_gdal_to_raster() {
 	CU_ASSERT(gddrv != NULL);
 	CU_ASSERT_EQUAL(destroy, 0);
 	CU_ASSERT(gdds != NULL);
-	CU_ASSERT_EQUAL(GDALGetRasterXSize(gdds), width);
-	CU_ASSERT_EQUAL(GDALGetRasterYSize(gdds), height);
+	CU_ASSERT_EQUAL((uint32_t)GDALGetRasterXSize(gdds), width);
+	CU_ASSERT_EQUAL((uint32_t)GDALGetRasterYSize(gdds), height);
 
 	rast = rt_raster_from_gdal_dataset(gdds);
 	CU_ASSERT(rast != NULL);
